"""add account table

Revision ID: 994170c26bc6
Revises: b6d49ed7cceb
Create Date: 2021-08-11 19:21:07.083253

"""
from typing import List, Tuple, Optional
import os
import json

from alembic import context, op
import sqlalchemy as sa
from sqlalchemy import orm
import inflection

from flexmeasures.data.models.user import Account, User
from flexmeasures.data.models.generic_assets import GenericAsset
from flexmeasures.data.models.time_series import Sensor
from flexmeasures.data.models.assets import Asset


# revision identifiers, used by Alembic.
revision = "994170c26bc6"
down_revision = "b6d49ed7cceb"
branch_labels = None
depends_on = None
asset_ownership_backup_script = "generic_asset_fm_user_ownership.sql"


def upgrade():
    """
    Add account table.
    1. Users need an account. You can pass this info in (user ID to account name) like this:
       flexmeasures db upgrade +1 -x '{"1": "One account", "2": "Bccount", "4": "Bccount"}'
       Note that user IDs are strings here, as this is a JSON array.
       The +1 makes sure we only upgrade by 1 revision, as these arguments are only meant to be used by this upgrade function.
       Users not mentioned here get an account derived from their email address' main domain, capitalized (info@company.com becomes "Company")
    2. The ownership of a generic_asset now goes to account.
       Here we fill in the user's new account (see point 1).
       (we save a backup of the generic_asset.owner_id info which linked to fm_user)
       The old-style asset's ownership remains in place for now! Our code will keep it consistent, until we have completed the move.
    """
    backup_generic_asset_user_associations()

    upgrade_schema()
    upgrade_data()

    op.alter_column("fm_user", "account_id", nullable=False)
    op.drop_column("generic_asset", "owner_id")


def downgrade():
    downgrade_schema()
    downgrade_data()


def upgrade_schema():
    op.create_table(
        "account",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("account_pkey")),
        sa.UniqueConstraint("name", name=op.f("account_name_key")),
    )
    op.add_column("fm_user", sa.Column("account_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        op.f("fm_user_account_id_account_fkey"),
        "fm_user",
        "account",
        ["account_id"],
        ["id"],
    )
    op.add_column("generic_asset", sa.Column("account_id", sa.Integer(), nullable=True))
    op.drop_constraint(
        "generic_asset_owner_id_fm_user_fkey", "generic_asset", type_="foreignkey"
    )
    op.create_foreign_key(
        op.f("generic_asset_account_id_account_fkey"),
        "generic_asset",
        "account",
        ["account_id"],
        ["id"],
        ondelete="CASCADE",
    )


def upgrade_data():
    # add custom accounts
    user_account_mappings = context.get_x_argument()
    connection = op.get_bind()
    session = orm.Session(bind=connection)
    for i, user_account_map in enumerate(user_account_mappings):
        print(user_account_map)
        user_account_dict = json.loads(user_account_map)
        for user_id, account_name in user_account_dict.items():
            print(
                f"Linking user {user_id} to account {account_name} (as from custom param) ..."
            )
            account = session.query(Account).filter_by(name=account_name).one_or_none()
            if account is None:
                print(f"need to create account {account_name} ...")
                account = Account(name=account_name)
                session.add(account)
                session.flush()
            user = session.query(User).filter_by(id=user_id).one_or_none()
            if not user:
                raise ValueError(f"User with ID {user_id} does not exist!")
            user.account_id = account.id

    # Make sure each existing user has an account
    for user in session.query(User).all():
        if user.account_id is None:
            domain = user.email.split("@")[-1].rsplit(".", maxsplit=1)[0]
            main_domain = domain.rsplit(".", maxsplit=1)[-1]
            account_name = inflection.titleize(main_domain)
            print(f"Linking user {user.id} to account {account_name} ...")
            account = session.query(Account).filter_by(name=account_name).one_or_none()
            if account is None:
                print(f"need to create account {account_name} ...")
                account = Account(name=account_name)
                session.add(account)
                session.flush()
            user.account_id = account.id

    # For all generic assets, set the user's account
    # We query the db for old ownership directly, as the generic asset code already points to account
    asset_ownership_db = _generic_asset_ownership()
    for generic_asset in session.query(GenericAsset).all():
        # 1. first look into GenericAsset ownership
        old_owner_id = _get_old_owner_id_from_db_result(
            asset_ownership_db, generic_asset.id
        )
        user = session.query(User).get(old_owner_id)
        # 2. Otherwise, then try the old-style Asset's ownership (via Sensor)
        if user is None:
            sensor = (
                session.query(Sensor)
                .filter_by(generic_asset_id=generic_asset.id)
                .one_or_none()
            )
            if sensor is None:
                raise ValueError(
                    f"GenericAsset {generic_asset.id} ({generic_asset.name}) does not have an assorted sensor. Please investigate ..."
                )
            asset = session.query(Asset).filter_by(id=sensor.id).one_or_none()
            if asset is None:
                print(
                    f"Generic asset {generic_asset.name} does not have an asset associated, probably because it's of type {generic_asset.generic_asset_type.name}."
                )
            else:
                user = asset.owner
        if user is not None:
            generic_asset.account_id = user.account.id
    session.commit()


def downgrade_schema():
    op.add_column(
        "generic_asset",
        sa.Column("owner_id", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.drop_constraint(
        op.f("generic_asset_account_id_account_fkey"),
        "generic_asset",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "generic_asset_owner_id_fm_user_fkey",
        "generic_asset",
        "fm_user",
        ["owner_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.drop_column("generic_asset", "account_id")
    op.drop_constraint(
        op.f("fm_user_account_id_account_fkey"), "fm_user", type_="foreignkey"
    )
    op.drop_column("fm_user", "account_id")
    op.drop_table("account")


def downgrade_data():
    if os.path.exists(asset_ownership_backup_script):
        print(
            f"Re-applying previous asset ownership from {asset_ownership_backup_script} ..."
        )
        connection = op.get_bind()
        session = orm.Session(bind=connection)
        with open(asset_ownership_backup_script, "r") as bckp_file:
            for statement in bckp_file.readlines():
                connection.execute(statement)
        session.commit()
    else:
        print(f"Could not find backup script {asset_ownership_backup_script} ...")
        print("Previous asset ownership information is probably lost.")


def backup_generic_asset_user_associations():
    asset_ownership_results = _generic_asset_ownership()
    backed_up_ownerships = 0
    with open(asset_ownership_backup_script, "w") as bckp_file:
        for aid, oid in asset_ownership_results:
            if oid is None:
                oid = "null"
            bckp_file.write(
                f"UPDATE generic_asset SET owner_id = {oid} WHERE id = {aid};\n"
            )
            backed_up_ownerships += 1

    if backed_up_ownerships > 0:
        print("Your generic_asset.owner_id associations are being dropped!")
        print(
            f"We saved UPDATE statements to put them back in {asset_ownership_backup_script}."
        )


def _generic_asset_ownership() -> List[Tuple[int, int]]:
    t_asset_owners = sa.Table(
        "generic_asset",
        sa.MetaData(),
        sa.Column("id", sa.Integer),
        sa.Column("owner_id", sa.Integer),
    )

    # Use SQLAlchemy's connection and transaction to go through the data
    connection = op.get_bind()

    # Select all existing ids that need migrating, while keeping names intact
    asset_ownership_results = connection.execute(
        sa.select(
            [
                t_asset_owners.c.id,
                t_asset_owners.c.owner_id,
            ]
        )
    ).fetchall()
    return asset_ownership_results


def _get_old_owner_id_from_db_result(
    generic_asset_ownership, asset_id
) -> Optional[int]:
    for aid, oid in generic_asset_ownership:
        if aid == asset_id:
            return oid
    return None



import os
import sys
import logging
from typing import Optional, List, Tuple
from datetime import datetime
from logging.config import dictConfig as loggingDictConfig
from pathlib import Path

from flask import Flask
from inflection import camelize

from flexmeasures.utils.config_defaults import (
    Config as DefaultConfig,
    required,
    warnable,
)


basedir = os.path.abspath(os.path.dirname(__file__))

flexmeasures_logging_config = {
    "version": 1,
    "formatters": {
        "default": {"format": "[FLEXMEASURES][%(asctime)s] %(levelname)s: %(message)s"},
        "detail": {
            "format": "[FLEXMEASURES][%(asctime)s] %(levelname)s: %(message)s [log made in %(pathname)s:%(lineno)d]"
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "stream": sys.stdout,
            "formatter": "default",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "INFO",
            "formatter": "detail",
            "filename": basedir + "/../../flexmeasures.log",
            "maxBytes": 10_000_000,
            "backupCount": 6,
        },
    },
    "root": {"level": "INFO", "handlers": ["console", "file"], "propagate": True},
}


def configure_logging():
    """Configure and register logging"""
    loggingDictConfig(flexmeasures_logging_config)


def read_config(app: Flask, custom_path_to_config: Optional[str]):
    """Read configuration from various expected sources, complain if not setup correctly. """

    if app.env not in (
        "documentation",
        "development",
        "testing",
        "staging",
        "production",
    ):
        print(
            'Flask(flexmeasures) environment needs to be either "documentation", "development", "testing", "staging" or "production".'
        )
        sys.exit(2)

    # First, load default config settings
    app.config.from_object(
        "flexmeasures.utils.config_defaults.%sConfig" % camelize(app.env)
    )

    # Now, potentially overwrite those from config file
    # These two locations are possible (besides the custom path)
    path_to_config_home = str(Path.home().joinpath(".flexmeasures.cfg"))
    path_to_config_instance = os.path.join(app.instance_path, "flexmeasures.cfg")
    if not app.testing:  # testing runs completely on defaults
        # If no custom path is given, this will try home dir first, then instance dir
        used_path_to_config = read_custom_config(
            app, custom_path_to_config, path_to_config_home, path_to_config_instance
        )

    # Check for missing values.
    # Documentation runs fine without them.
    if not app.testing and app.env != "documentation":
        if not are_required_settings_complete(app):
            if not os.path.exists(used_path_to_config):
                print(
                    f"You can provide these settings ― as environment variables or in your config file (e.g. {path_to_config_home} or {path_to_config_instance})."
                )
            else:
                print(
                    f"Please provide these settings ― as environment variables or in your config file ({used_path_to_config})."
                )
            sys.exit(2)
        missing_fields, config_warnings = get_config_warnings(app)
        if len(config_warnings) > 0:
            for warning in config_warnings:
                print(f"Warning: {warning}")
            print(f"You might consider setting {', '.join(missing_fields)}.")

    # Set the desired logging level on the root logger (controlling extension logging level)
    # and this app's logger.
    logging.getLogger().setLevel(app.config.get("LOGGING_LEVEL", "INFO"))
    app.logger.setLevel(app.config.get("LOGGING_LEVEL", "INFO"))
    # print("Logging level is %s" % logging.getLevelName(app.logger.level))

    app.config["START_TIME"] = datetime.utcnow()


def read_custom_config(
    app, suggested_path_to_config, path_to_config_home, path_to_config_instance
) -> str:
    """ read in a custom config file or env vars. Return the path to the config file."""
    if suggested_path_to_config is not None and not os.path.exists(
        suggested_path_to_config
    ):
        print(f"Cannot find config file {suggested_path_to_config}!")
        sys.exit(2)
    if suggested_path_to_config is None:
        path_to_config = path_to_config_home
        if not os.path.exists(path_to_config):
            path_to_config = path_to_config_instance
    else:
        path_to_config = suggested_path_to_config
    try:
        app.config.from_pyfile(path_to_config)
    except FileNotFoundError:
        pass
    # Finally, all required variables can be set as env var:
    for req_var in required:
        app.config[req_var] = os.getenv(req_var, app.config.get(req_var, None))
    return path_to_config


def are_required_settings_complete(app) -> bool:
    """
    Check if all settings we expect are not None. Return False if they are not.
    Printout helpful advice.
    """
    expected_settings = [s for s in get_configuration_keys(app) if s in required]
    missing_settings = [s for s in expected_settings if app.config.get(s) is None]
    if len(missing_settings) > 0:
        print(
            f"Missing the required configuration settings: {', '.join(missing_settings)}"
        )
        return False
    return True


def get_config_warnings(app) -> Tuple[List[str], List[str]]:
    """return missing settings and the warnings for them."""
    missing_settings = []
    config_warnings = []
    for setting, warning in warnable.items():
        if app.config.get(setting) is None:
            missing_settings.append(setting)
            config_warnings.append(warning)
    config_warnings = list(set(config_warnings))
    return missing_settings, config_warnings


def get_configuration_keys(app) -> List[str]:
    """
    Collect all members of DefaultConfig who are not in-built fields or callables.
    """
    return [
        a
        for a in DefaultConfig.__dict__
        if not a.startswith("__") and not callable(getattr(DefaultConfig, a))
    ]



from functools import wraps

from flask import current_app, abort
from marshmallow import fields
from sqlalchemy.exc import IntegrityError
from webargs.flaskparser import use_args
from flask_security import current_user
from flask_security.recoverable import send_reset_password_instructions
from flask_json import as_json

from flexmeasures.data.models.user import User as UserModel
from flexmeasures.data.schemas.users import UserSchema
from flexmeasures.data.services.users import (
    get_users,
    set_random_password,
    remove_cookie_and_token_access,
)
from flexmeasures.data.auth_setup import unauthorized_handler
from flexmeasures.api.common.responses import required_info_missing
from flexmeasures.data.config import db

"""
API endpoints to manage users.

Both POST (to create) and DELETE are not accessible via the API, but as CLI functions.
"""

user_schema = UserSchema()
users_schema = UserSchema(many=True)


@use_args({"include_inactive": fields.Bool(missing=False)}, location="query")
@as_json
def get(args):
    """List all users."""
    users = get_users(only_active=not args["include_inactive"])
    return users_schema.dump(users), 200


def load_user(admins_only: bool = False):
    """Decorator which loads a user by the Id expected in the path.
    Raises 400 if that is not possible due to wrong parameters.
    Raises 404 if user is not found.
    Raises 403 if unauthorized:
    Only the user themselves or admins can access a user object.
    The admins_only parameter can be used if not even the user themselves
    should be allowed.

        @app.route('/user/<id>')
        @check_user
        def get_user(user):
            return user_schema.dump(user), 200

    The route must specify one parameter ― id.
    """

    def wrapper(fn):
        @wraps(fn)
        @as_json
        def decorated_endpoint(*args, **kwargs):

            args = list(args)
            if len(args) == 0:
                current_app.logger.warning("Request missing id.")
                return required_info_missing(["id"])
            if len(args) > 1:
                return (
                    dict(
                        status="UNEXPECTED_PARAMS",
                        message="Only expected one parameter (id).",
                    ),
                    400,
                )

            try:
                id = int(args[0])
            except ValueError:
                current_app.logger.warning("Cannot parse ID argument from request.")
                return required_info_missing(["id"], "Cannot parse ID arg as int.")

            user: UserModel = UserModel.query.filter_by(id=int(id)).one_or_none()

            if user is None:
                raise abort(404, f"User {id} not found")

            if not current_user.has_role("admin"):
                if admins_only or user != current_user:
                    return unauthorized_handler(None, [])

            args = (user,)
            return fn(*args, **kwargs)

        return decorated_endpoint

    return wrapper


@load_user()
@as_json
def fetch_one(user: UserModel):
    """Fetch a given user"""
    return user_schema.dump(user), 200


@load_user()
@use_args(UserSchema(partial=True))
@as_json
def patch(db_user: UserModel, user_data: dict):
    """Update a user given its identifier"""
    allowed_fields = ["email", "username", "active", "timezone", "flexmeasures_roles"]
    for k, v in [(k, v) for k, v in user_data.items() if k in allowed_fields]:
        # Don't allow users who edit themselves to edit sensitive fields
        if current_user.id == db_user.id and k in ("active", "flexmeasures_roles"):
            return unauthorized_handler(None, [])
        setattr(db_user, k, v)
        if k == "active" and v is False:
            remove_cookie_and_token_access(db_user)
    db.session.add(db_user)
    try:
        db.session.commit()
    except IntegrityError as ie:
        return dict(message="Duplicate user already exists", detail=ie._message()), 400
    return user_schema.dump(db_user), 200


@load_user(admins_only=True)
@as_json
def reset_password(user):
    """
    Reset the user's current password, cookies and auth tokens.
    Send a password reset link to the user.
    """
    set_random_password(user)
    remove_cookie_and_token_access(user)
    send_reset_password_instructions(user)

    # commit only if sending instructions worked, as well
    db.session.commit()



from flexmeasures.data.models.time_series import Sensor


def make_sensor_data_request(num_values: int = 6, duration: str = "PT1H") -> dict:
    sensor = Sensor.query.filter(Sensor.name == "some gas sensor").one_or_none()
    message: dict = {
        "type": "PostSensorDataRequest",
        "sensor": f"ea1.2021-01.io.flexmeasures:fm1.{sensor.id}",
        "values": num_values * [-11.28],
        "start": "2021-06-07T00:00:00+02:00",
        "duration": duration,
        "unit": "m³/h",
    }
    if num_values == 1:
        # flatten [<float>] to <float>
        message["values"] = message["values"][0]
    return message



#!/usr/bin/env python

import os
from typing import Tuple, List, Dict
import json
from datetime import datetime

import click
from flask import Flask, current_app
import requests
import pytz

from flexmeasures.utils.time_utils import as_server_time, get_timezone
from flexmeasures.utils.geo_utils import compute_irradiance
from flexmeasures.data.services.resources import find_closest_weather_sensor
from flexmeasures.data.config import db
from flexmeasures.data.transactional import task_with_status_report
from flexmeasures.data.models.weather import Weather
from flexmeasures.data.models.data_sources import DataSource

FILE_PATH_LOCATION = "/../raw_data/weather-forecasts"
DATA_SOURCE_NAME = "OpenWeatherMap"


class LatLngGrid(object):
    """
    Represents a grid in latitude and longitude notation for some rectangular region of interest (ROI).
    The specs are a top-left and a bottom-right coordinate, as well as the number of cells in both directions.
    The class provides two ways of conceptualising cells which nicely cover the grid: square cells and hexagonal cells.
    For both, locations can be computed which represent the corners of said cells.
    Examples:
          - 4 cells in square: 9 unique locations in a 2x2 grid (4*4 locations, of which 7 are covered by another cell)
          - 4 cells in hex: 13 unique locations in a 2x2 grid (4*6 locations, of which 11 are already covered)
          - 10 cells in square: 18 unique locations in a 5x2 grid (10*4 locations, of which 11 are already covered)
          - 10 cells in hex: 34 unique locations in a 5x2 grid (10*6 locations, of which 26 are already covered)
    The top-right and bottom-left locations are always at the center of a cell, unless the grid has 1 row or 1 column.
    In those case, these locations are closer to one side of the cell.
    """

    top_left: Tuple[float, float]
    bottom_right: Tuple[float, float]
    num_cells_lat: int
    num_cells_lng: int

    def __init__(
        self,
        top_left: Tuple[float, float],
        bottom_right: Tuple[float, float],
        num_cells_lat: int,
        num_cells_lng: int,
    ):
        self.top_left = top_left
        self.bottom_right = bottom_right
        self.num_cells_lat = num_cells_lat
        self.num_cells_lng = num_cells_lng
        self.cell_size_lat = self.compute_cell_size_lat()
        self.cell_size_lng = self.compute_cell_size_lng()

        # correct top-left and bottom-right if there is only one cell
        if self.num_cells_lat == 1:  # if only one row
            self.top_left = (
                self.top_left[0] + self.cell_size_lat / 4,
                self.top_left[1],
            )
            self.bottom_right = (
                self.bottom_right[0] - self.cell_size_lat / 4,
                self.bottom_right[1],
            )
        if self.num_cells_lng == 1:
            self.top_left = (
                self.top_left[0],
                self.top_left[1] + self.cell_size_lng / 4,
            )
            self.bottom_right = (
                self.bottom_right[0],
                self.bottom_right[1] - self.cell_size_lng / 4,
            )

    def __repr__(self) -> str:
        return (
            f"<LatLngGrid top-left:{self.top_left}, bot-right:{self.bottom_right},"
            + f"num_lat:{self.num_cells_lat}, num_lng:{self.num_cells_lng}>"
        )

    def get_locations(self, method: str) -> List[Tuple[float, float]]:
        """Get locations by method ("square" or "hex")"""
        click.echo(self)
        click.echo()

        if method == "hex":
            locations = self.locations_hex()
            click.echo(
                "[FLEXMEASURES] Number of locations in hex grid: "
                + str(len(self.locations_hex()))
            )
            return locations
        elif method == "square":
            locations = self.locations_square()
            click.echo(
                "[FLEXMEASURES] Number of locations in square grid: "
                + str(len(self.locations_square()))
            )
            return locations
        else:
            raise Exception(
                "Method must either be 'square' or 'hex'! (is: %s)" % method
            )

    def compute_cell_size_lat(self) -> float:
        """Calculate the step size between latitudes"""
        if self.num_cells_lat != 1:
            return (self.bottom_right[0] - self.top_left[0]) / (self.num_cells_lat - 1)
        else:
            return (self.bottom_right[0] - self.top_left[0]) * 2

    def compute_cell_size_lng(self) -> float:
        """Calculate the step size between longitudes"""
        if self.num_cells_lng != 1:
            return (self.bottom_right[1] - self.top_left[1]) / (self.num_cells_lng - 1)
        else:
            return (self.bottom_right[1] - self.top_left[1]) * 2

    def locations_square(self) -> List[Tuple[float, float]]:
        """square pattern"""
        locations = []

        # For each odd cell row, add all the coordinates of the row's cells
        for ilat in range(0, self.num_cells_lat, 2):
            lat = self.top_left[0] + ilat * self.cell_size_lat
            for ilng in range(self.num_cells_lng):
                lng = self.top_left[1] + ilng * self.cell_size_lng
                nw = (
                    lat - self.cell_size_lat / 2,
                    lng - self.cell_size_lng / 2,
                )  # North west coordinate of the cell
                locations.append(nw)
                sw = (
                    lat + self.cell_size_lat / 2,
                    lng - self.cell_size_lng / 2,
                )  # South west coordinate
                locations.append(sw)
            ne = (
                lat - self.cell_size_lat / 2,
                lng + self.cell_size_lng / 2,
            )  # North east coordinate
            locations.append(ne)
            se = (
                lat + self.cell_size_lat / 2,
                lng + self.cell_size_lng / 2,
            )  # South east coordinate
            locations.append(se)

        # In case of an even number of cell rows, add the southern coordinates of the southern most row
        if not self.num_cells_lat % 2:
            lat = self.top_left[0] + (self.num_cells_lat - 1) * self.cell_size_lat
            for ilng in range(self.num_cells_lng):
                lng = self.top_left[1] + ilng * self.cell_size_lng
                sw = (
                    lat + self.cell_size_lat / 2,
                    lng - self.cell_size_lng / 2,
                )  # South west coordinate
                locations.append(sw)
            se = (
                lat + self.cell_size_lat / 2,
                lng + self.cell_size_lng / 2,
            )  # South east coordinate
            locations.append(se)

        return locations

    def locations_hex(self) -> List[Tuple[float, float]]:
        """The hexagonal pattern - actually leaves out one cell for every even row."""
        locations = []

        # For each odd cell row, add all the coordinates of the row's cells
        for ilat in range(0, self.num_cells_lat, 2):
            lat = self.top_left[0] + ilat * self.cell_size_lat
            for ilng in range(self.num_cells_lng):
                lng = self.top_left[1] + ilng * self.cell_size_lng
                n = (
                    lat - self.cell_size_lat * 2 / 3,
                    lng,
                )  # North coordinate of the cell
                locations.append(n)
                nw = (
                    lat - self.cell_size_lat * 1 / 4,
                    lng - self.cell_size_lng * 1 / 2,
                )  # North west coordinate
                locations.append(nw)
                s = (lat + self.cell_size_lat * 2 / 3, lng)  # South coordinate
                locations.append(s)
                sw = (
                    lat + self.cell_size_lat * 1 / 4,
                    lng - self.cell_size_lng * 1 / 2,
                )  # South west coordinate
                locations.append(sw)
            ne = (
                lat - self.cell_size_lat * 1 / 4,
                lng + self.cell_size_lng * 1 / 2,
            )  # North east coordinate
            locations.append(ne)
            se = (
                lat + self.cell_size_lat * 1 / 4,
                lng + self.cell_size_lng * 1 / 2,
            )  # South east coordinate
            locations.append(se)

        # In case of an even number of cell rows, add the southern coordinates of the southern most row
        if not self.num_cells_lat % 2:
            lat = self.top_left[0] + (self.num_cells_lat - 1) * self.cell_size_lat
            for ilng in range(
                self.num_cells_lng - 1
            ):  # One less cell in even rows of hexagonal locations
                # Cells are shifted half a cell to the right in even rows of hex locations
                lng = self.top_left[1] + (ilng + 1 / 2) * self.cell_size_lng
                s = (lat + self.cell_size_lat / 3 ** (1 / 2), lng)  # South coordinate
                locations.append(s)
                sw = (
                    lat + self.cell_size_lat / 2,
                    lng - self.cell_size_lat / 3 ** (1 / 2) / 2,
                )  # South west coordinates
                locations.append(sw)
                se = (
                    lat + self.cell_size_lat / 2,
                    lng + self.cell_size_lng / 3 ** (1 / 2) / 2,
                )  # South east coordinates
                locations.append(se)
        return locations


def get_cell_nums(
    tl: Tuple[float, float], br: Tuple[float, float], num_cells: int = 9
) -> Tuple[int, int]:
    """
    Compute the number of cells in both directions, latitude and longitude.
    By default, a square grid with N=9 cells is computed, so 3 by 3.
    For N with non-integer square root, the function will determine a nice cell pattern.
    :param tl: top-left (lat, lng) tuple of ROI
    :param br: bottom-right (lat, lng) tuple of ROI
    :param num_cells: number of cells (9 by default, leading to a 3x3 grid)
    """

    def factors(n):
        """Factors of a number n"""
        return set(
            factor
            for i in range(1, int(n ** 0.5) + 1)
            if n % i == 0
            for factor in (i, n // i)
        )

    # Find closest integers n1 and n2 that, when multiplied, equal n
    n1 = min(factors(num_cells), key=lambda x: abs(x - num_cells ** (1 / 2)))
    n2 = num_cells // n1

    # Assign largest integer to lat or lng depending on which has the largest spread (earth curvature neglected)
    if br[0] - tl[0] > br[1] - tl[1]:
        return max(n1, n2), min(n1, n2)
    else:
        return min(n1, n2), max(n1, n2)


def get_region_from_assets() -> Tuple[Tuple[float, float], Tuple[float, float]]:
    """
    Create a suitable region of interest from all asset locations.
    Currently not used. (we simply pass top left and bottom right to this script).
    Should in any case later probably contact the database actually.
    """
    assets_path = "../raw_data/assets.json"
    lats, lngs = [], []
    if os.path.exists(assets_path):
        with open(assets_path, "r") as json_data:
            assets = json.load(json_data)
    else:
        raise Exception("File not found: %s" % assets_path)

    for asset in assets:
        if "latitude" in asset and "longitude" in asset:
            lats.append(asset["latitude"])
            lngs.append(asset["longitude"])
        else:
            click.echo(
                "[FLEXMEASURES] Asset %s has no latitude and/or longitude."
                % asset["name"]
            )
    top_left = min(lats), min(lngs)
    bottom_right = max(lats), max(lngs)
    return top_left, bottom_right


def make_file_path(app: Flask, region: str) -> str:
    """Ensure and return path for weather data"""
    data_path = app.root_path + FILE_PATH_LOCATION
    if not os.path.exists(data_path):
        if os.path.exists(app.root_path + "/../raw_data"):
            click.echo("[FLEXMEASURES] Creating %s ..." % data_path)
            os.mkdir(data_path)
        else:
            raise Exception("No %s/../raw_data directory found." % app.root_path)
    # optional: extend with subpath for region
    if region is not None and region != "":
        region_data_path = "%s/%s" % (data_path, region)
        if not os.path.exists(region_data_path):
            click.echo("[FLEXMEASURES] Creating %s ..." % region_data_path)
            os.mkdir(region_data_path)
        data_path = region_data_path
    return data_path


def get_data_source() -> DataSource:
    """Make sure we have a data source"""
    data_source = DataSource.query.filter_by(
        name=DATA_SOURCE_NAME, type="forecasting script"
    ).one_or_none()
    if data_source is None:
        data_source = DataSource(name=DATA_SOURCE_NAME, type="forecasting script")
        db.session.add(data_source)
    return data_source


def call_openweatherapi(
    api_key: str, location: Tuple[float, float]
) -> Tuple[int, List[Dict]]:
    """
    Make a single "one-call" to the Open Weather API and return the API timestamp as well as the 48 hourly forecasts.
    See https://openweathermap.org/api/one-call-api for docs.
    Note that the first forecast is about the current hour.
    """
    query_str = f"lat={location[0]}&lon={location[1]}&units=metric&exclude=minutely,daily,alerts&appid={api_key}"
    res = requests.get(f"http://api.openweathermap.org/data/2.5/onecall?{query_str}")
    assert (
        res.status_code == 200
    ), f"OpenWeatherMap returned status code {res.status_code}: {res.text}"
    data = res.json()
    return data["current"]["dt"], data["hourly"]


def save_forecasts_in_db(
    api_key: str,
    locations: List[Tuple[float, float]],
    data_source: DataSource,
    max_degree_difference_for_nearest_weather_sensor: int = 2,
):
    """Process the response from OpenWeatherMap API into Weather timed values.
    Collects all forecasts for all locations and all sensors at all locations, then bulk-saves them.
    """
    click.echo("[FLEXMEASURES] Getting weather forecasts:")
    click.echo("[FLEXMEASURES]  Latitude, Longitude")
    click.echo("[FLEXMEASURES] -----------------------")
    db_forecasts = []
    weather_sensors: dict = {}  # keep track of the sensors to save lookups

    for location in locations:
        click.echo("[FLEXMEASURES] %s, %s" % location)

        api_timestamp, forecasts = call_openweatherapi(api_key, location)
        time_of_api_call = as_server_time(
            datetime.fromtimestamp(api_timestamp, tz=get_timezone())
        ).replace(second=0, microsecond=0)
        click.echo(
            "[FLEXMEASURES] Called OpenWeatherMap API successfully at %s."
            % time_of_api_call
        )

        # map sensor name in our db to sensor name/label in OWM response
        sensor_name_mapping = dict(
            temperature="temp", wind_speed="wind_speed", radiation="clouds"
        )

        # loop through forecasts, including the one of current hour (horizon 0)
        for fc in forecasts:
            fc_datetime = as_server_time(
                datetime.fromtimestamp(fc["dt"], get_timezone())
            ).replace(second=0, microsecond=0)
            fc_horizon = fc_datetime - time_of_api_call
            click.echo(
                "[FLEXMEASURES] Processing forecast for %s (horizon: %s) ..."
                % (fc_datetime, fc_horizon)
            )
            for flexmeasures_sensor_type in sensor_name_mapping.keys():
                needed_response_label = sensor_name_mapping[flexmeasures_sensor_type]
                if needed_response_label in fc:
                    weather_sensor = weather_sensors.get(flexmeasures_sensor_type, None)
                    if weather_sensor is None:
                        weather_sensor = find_closest_weather_sensor(
                            flexmeasures_sensor_type, lat=location[0], lng=location[1]
                        )
                        if weather_sensor is not None:
                            # Complain if the nearest weather sensor is further away than 2 degrees
                            if abs(
                                location[0] - weather_sensor.latitude
                            ) > max_degree_difference_for_nearest_weather_sensor or abs(
                                location[1] - weather_sensor.longitude
                                > max_degree_difference_for_nearest_weather_sensor
                            ):
                                raise Exception(
                                    f"No sufficiently close weather sensor found (within 2 degrees distance) for type {flexmeasures_sensor_type}! We're looking for: {location}, closest available: ({weather_sensor.latitude}, {weather_sensor.longitude})"
                                )
                            weather_sensors[flexmeasures_sensor_type] = weather_sensor
                        else:
                            raise Exception(
                                "No weather sensor set up for this sensor type (%s)"
                                % flexmeasures_sensor_type
                            )

                    fc_value = fc[needed_response_label]
                    # the radiation is not available in OWM -> we compute it ourselves
                    if flexmeasures_sensor_type == "radiation":
                        fc_value = compute_irradiance(
                            location[0],
                            location[1],
                            fc_datetime,
                            # OWM sends cloud coverage in percent, we need a ratio
                            fc[needed_response_label] / 100.0,
                        )

                    db_forecasts.append(
                        Weather(
                            datetime=fc_datetime,
                            horizon=fc_horizon,
                            value=fc_value,
                            sensor_id=weather_sensor.id,
                            data_source_id=data_source.id,
                        )
                    )
                else:
                    # we will not fail here, but issue a warning
                    msg = "No label '%s' in response data for time %s" % (
                        needed_response_label,
                        fc_datetime,
                    )
                    click.echo("[FLEXMEASURES] %s" % msg)
                    current_app.logger.warning(msg)
    if len(db_forecasts) == 0:
        # This is probably a serious problem
        raise Exception(
            "Nothing to put in the database was produced. That does not seem right..."
        )
    db.session.bulk_save_objects(db_forecasts)


def save_forecasts_as_json(
    api_key: str, locations: List[Tuple[float, float]], data_path: str
):
    """Get forecasts, then store each as a JSON file"""
    click.echo("[FLEXMEASURES] Getting weather forecasts:")
    click.echo("[FLEXMEASURES]  Latitude, Longitude")
    click.echo("[FLEXMEASURES]  ----------------------")
    for location in locations:
        click.echo("[FLEXMEASURES] %s, %s" % location)
        api_timestamp, forecasts = call_openweatherapi(api_key, location)
        time_of_api_call = as_server_time(
            datetime.fromtimestamp(api_timestamp, tz=pytz.utc)
        ).replace(second=0, microsecond=0)
        now_str = time_of_api_call.strftime("%Y-%m-%dT%H-%M-%S")
        path_to_files = os.path.join(data_path, now_str)
        if not os.path.exists(path_to_files):
            click.echo(f"Making directory: {path_to_files} ...")
            os.mkdir(path_to_files)
        forecasts_file = "%s/forecast_lat_%s_lng_%s.json" % (
            path_to_files,
            str(location[0]),
            str(location[1]),
        )
        with open(forecasts_file, "w") as outfile:
            json.dump(forecasts, outfile)


@task_with_status_report
def get_weather_forecasts(
    app: Flask,
    region: str,
    location: str,
    num_cells: int,
    method: str,
    store_in_db: bool,
):
    """
    Get current weather forecasts for a latitude/longitude grid and store them in individual json files.
    Note that 1000 free calls per day can be made to the OpenWeatherMap API,
    so we can make a call every 15 minutes for up to 10 assets or every hour for up to 40 assets (or get a paid account).
    """
    if app.config.get("OPENWEATHERMAP_API_KEY") is None:
        raise Exception("Setting OPENWEATHERMAP_API_KEY not available.")

    if (
        location.count(",") == 0
        or location.count(",") != location.count(":") + 1
        or location.count(":") == 1
        and (
            location.find(",") > location.find(":")
            or location.find(",", location.find(",") + 1) < location.find(":")
        )
    ):
        raise Exception(
            'location parameter "%s" seems malformed. Please use "latitude,longitude" or '
            ' "top-left-latitude,top-left-longitude:bottom-right-latitude,bottom-right-longitude"'
            % location
        )

    location_identifiers = tuple(location.split(":"))

    if len(location_identifiers) == 1:
        locations = [tuple(float(s) for s in location_identifiers[0].split(","))]
        click.echo("[FLEXMEASURES] Only one location: %s,%s." % locations[0])
    elif len(location_identifiers) == 2:
        click.echo(
            "[FLEXMEASURES] Making a grid of locations between top/left %s and bottom/right %s ..."
            % location_identifiers
        )
        top_left = tuple(float(s) for s in location_identifiers[0].split(","))
        if len(top_left) != 2:
            raise Exception(
                "top-left parameter '%s' is invalid." % location_identifiers[0]
            )
        bottom_right = tuple(float(s) for s in location_identifiers[1].split(","))
        if len(bottom_right) != 2:
            raise Exception(
                "bottom-right parameter '%s' is invalid." % location_identifiers[1]
            )

        num_lat, num_lng = get_cell_nums(top_left, bottom_right, num_cells)

        locations = LatLngGrid(
            top_left=top_left,
            bottom_right=bottom_right,
            num_cells_lat=num_lat,
            num_cells_lng=num_lng,
        ).get_locations(method)
    else:
        raise Exception("location parameter '%s' has too many locations." % location)

    api_key = app.config.get("OPENWEATHERMAP_API_KEY")

    # Save the results
    if store_in_db:
        save_forecasts_in_db(api_key, locations, data_source=get_data_source())
    else:
        save_forecasts_as_json(
            api_key, locations, data_path=make_file_path(app, region)
        )



# Time resolutions
resolutions = ["15T", "1h", "1d", "1w"]

# The confidence interval for forecasting
confidence_interval_width = 0.9

naming_convention = {
    "ix": "%(column_0_label)s_idx",
    "uq": "%(table_name)s_%(column_0_name)s_key",
    "ck": "%(table_name)s_%(constraint_name)s_ck",
    "fk": "%(table_name)s_%(column_0_name)s_%(referred_table_name)s_fkey",
    "pk": "%(table_name)s_pkey",
}


class ModelException(Exception):
    pass



"""default-resolution-for-existing-sensors

Revision ID: 550a9020f1bf
Revises: a5b970eadb3b
Create Date: 2020-11-05 17:48:49.670289

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "550a9020f1bf"
down_revision = "a5b970eadb3b"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "UPDATE asset SET event_resolution = interval '00:15:00' where event_resolution = interval '00:00:00';"
    )
    op.execute(
        "UPDATE weather_sensor SET event_resolution = interval '00:15:00' where event_resolution = interval '00:00:00';"
    )
    op.execute(
        "UPDATE market SET event_resolution = interval '00:15:00' where event_resolution = interval '00:00:00';"
    )


def downgrade():
    op.execute(
        "UPDATE asset SET event_resolution = interval '00:00:00' where event_resolution = interval '00:15:00';"
    )
    op.execute(
        "UPDATE weather_sensor SET event_resolution = interval '00:00:00' where event_resolution = interval '00:15:00';"
    )
    op.execute(
        "UPDATE market SET event_resolution = interval '00:00:00' where event_resolution = interval '00:15:00';"
    )



from altair.utils.html import spec_to_html
from flask import current_app
from flask_classful import FlaskView, route
from flask_security import login_required, roles_required
from marshmallow import fields
from webargs.flaskparser import use_kwargs

from flexmeasures.data.schemas.times import AwareDateTimeField
from flexmeasures.api.dev.sensors import SensorAPI
from flexmeasures.ui.utils.view_utils import render_flexmeasures_template


class SensorUI(FlaskView):
    """
    This view creates several new UI endpoints for viewing sensors.

    todo: consider extending this view for crud purposes
    """

    route_base = "/sensors"

    @login_required
    @roles_required("admin")  # todo: remove after we check for sensor ownership
    @route("/<id>/chart/")
    @use_kwargs(
        {
            "event_starts_after": AwareDateTimeField(format="iso", required=False),
            "event_ends_before": AwareDateTimeField(format="iso", required=False),
            "beliefs_after": AwareDateTimeField(format="iso", required=False),
            "beliefs_before": AwareDateTimeField(format="iso", required=False),
            "dataset_name": fields.Str(required=False),
        },
        location="query",
    )
    def get_chart(self, id, **kwargs):
        """GET from /sensors/<id>/chart"""
        chart_specs = SensorAPI().get_chart(
            id, include_data=True, as_html=True, **kwargs
        )
        return spec_to_html(
            chart_specs,
            "vega-lite",
            vega_version=current_app.config.get("FLEXMEASURES_JS_VERSIONS").vega,
            vegaembed_version=current_app.config.get(
                "FLEXMEASURES_JS_VERSIONS"
            ).vegaembed,
            vegalite_version=current_app.config.get(
                "FLEXMEASURES_JS_VERSIONS"
            ).vegalite,
        )

    @login_required
    @roles_required("admin")  # todo: remove after we check for sensor ownership
    def get(self, id: int):
        """GET from /sensors/<id>"""
        return render_flexmeasures_template(
            "views/sensors.html",
            sensor_id=id,
            msg="",
        )



# flake8: noqa: E402
import os
import time
from typing import Optional

from flask import Flask, g, request
from flask.cli import load_dotenv
from flask_mail import Mail
from flask_sslify import SSLify
from flask_json import FlaskJSON
from flask_cors import CORS

from redis import Redis
from rq import Queue


def create(env: Optional[str] = None, path_to_config: Optional[str] = None) -> Flask:
    """
    Create a Flask app and configure it.

    Set the environment by setting FLASK_ENV as environment variable (also possible in .env).
    Or, overwrite any FLASK_ENV setting by passing an env in directly (useful for testing for instance).

    A path to a config file can be passed in (otherwise a config file will be searched in the home or instance directories)
    """

    from flexmeasures.utils import config_defaults
    from flexmeasures.utils.config_utils import read_config, configure_logging
    from flexmeasures.utils.app_utils import set_secret_key, init_sentry
    from flexmeasures.utils.error_utils import add_basic_error_handlers

    # Create app

    configure_logging()  # do this first, see http://flask.pocoo.org/docs/dev/logging/
    # we're loading dotenv files manually & early (can do Flask.run(load_dotenv=False)),
    # as we need to know the ENV now (for it to be recognised by Flask()).
    load_dotenv()
    app = Flask("flexmeasures")

    if env is not None:  # overwrite
        app.env = env
    if app.env == "testing":
        app.testing = True
    if app.env == "development":
        app.debug = config_defaults.DevelopmentConfig.DEBUG

    # App configuration

    read_config(app, custom_path_to_config=path_to_config)
    add_basic_error_handlers(app)
    if not app.env in ("development", "documentation") and not app.testing:
        init_sentry(app)

    app.mail = Mail(app)
    FlaskJSON(app)
    cors = CORS(app)

    # configure Redis (for redis queue)
    if app.testing:
        from fakeredis import FakeStrictRedis

        app.queues = dict(
            forecasting=Queue(connection=FakeStrictRedis(), name="forecasting"),
            scheduling=Queue(connection=FakeStrictRedis(), name="scheduling"),
        )
    else:
        redis_conn = Redis(
            app.config["FLEXMEASURES_REDIS_URL"],
            port=app.config["FLEXMEASURES_REDIS_PORT"],
            db=app.config["FLEXMEASURES_REDIS_DB_NR"],
            password=app.config["FLEXMEASURES_REDIS_PASSWORD"],
        )
        """ FWIW, you could use redislite like this (not on non-recent os.name=="nt" systems or PA, sadly):
            from redislite import Redis
            redis_conn = Redis("MY-DB-NAME", unix_socket_path="/tmp/my-redis.socket",
            )
        """
        app.queues = dict(
            forecasting=Queue(connection=redis_conn, name="forecasting"),
            scheduling=Queue(connection=redis_conn, name="scheduling"),
        )

    # Some basic security measures

    if not app.env == "documentation":
        set_secret_key(app)
        if app.config.get("SECURITY_PASSWORD_SALT", None) is None:
            app.config["SECURITY_PASSWORD_SALT"] = app.config["SECRET_KEY"]
    else:
        app.config["SECRET_KEY"] = "dummy-secret-for-documentation-creation"
    if not app.env in ("documentation", "development"):
        SSLify(app)

    # Register database and models, including user auth security handlers

    from flexmeasures.data import register_at as register_db_at

    register_db_at(app)

    # Register the API

    from flexmeasures.api import register_at as register_api_at

    register_api_at(app)

    # Register plugins
    # If plugins register routes, they'll have precedence over standard UI
    # routes (first registration wins). However, we want to control "/" separately.

    from flexmeasures.utils.app_utils import root_dispatcher, register_plugins

    app.add_url_rule("/", view_func=root_dispatcher)
    register_plugins(app)

    # Register the UI

    from flexmeasures.ui import register_at as register_ui_at

    register_ui_at(app)

    # Profile endpoints (if needed, e.g. during development)

    @app.before_request
    def before_request():
        if app.config.get("FLEXMEASURES_PROFILE_REQUESTS", False):
            g.start = time.time()

    @app.teardown_request
    def teardown_request(exception=None):
        if app.config.get("FLEXMEASURES_PROFILE_REQUESTS", False):
            diff = time.time() - g.start
            if all([kw not in request.url for kw in ["/static", "favicon.ico"]]):
                app.logger.info(
                    f"[PROFILE] {str(round(diff, 2)).rjust(6)} seconds to serve {request.url}."
                )

    return app



from flask import url_for

import pytest

from flexmeasures.data.services.users import find_user_by_email
from flexmeasures.ui.tests.utils import mock_asset_response, mock_api_data_as_form_input

"""
Testing if our asset UI proceeds with the expected roundtrip.
Here, we mock the API responses (we have to, as our UI layer contacts FlexMeasures as a server, which does not run during tests).
The real logic tests are done in the api package, which is also the better place for that.
"""


def test_assets_page_empty(db, client, requests_mock, as_prosumer):
    requests_mock.get("http://localhost//api/v2_0/assets", status_code=200, json={})
    asset_index = client.get(url_for("AssetCrudUI:index"), follow_redirects=True)
    assert asset_index.status_code == 200


@pytest.mark.parametrize("use_owned_by", [False, True])
def test_assets_page_nonempty(db, client, requests_mock, as_prosumer, use_owned_by):
    mock_assets = mock_asset_response(multiple=True)
    requests_mock.get(
        "http://localhost//api/v2_0/assets", status_code=200, json=mock_assets
    )
    if use_owned_by:
        asset_index = client.get(
            url_for("AssetCrudUI:owned_by", owner_id=mock_assets[0]["owner_id"])
        )
    else:
        asset_index = client.get(url_for("AssetCrudUI:index"))
    for asset in mock_assets:
        assert asset["display_name"].encode() in asset_index.data


def test_new_asset_page(client, setup_assets, as_admin):
    asset_page = client.get(url_for("AssetCrudUI:get", id="new"), follow_redirects=True)
    assert asset_page.status_code == 200
    assert b"Creating a new asset" in asset_page.data


def test_asset_page(db, client, setup_assets, requests_mock, as_prosumer):
    prosumer = find_user_by_email("test_prosumer@seita.nl")
    asset = prosumer.assets[0]
    db.session.expunge(prosumer)
    mock_asset = mock_asset_response(as_list=False)
    mock_asset["capacity_in_mw"] = asset.capacity_in_mw
    mock_asset["latitude"] = asset.latitude
    mock_asset["longitude"] = asset.longitude

    requests_mock.get(
        f"http://localhost//api/v2_0/asset/{asset.id}", status_code=200, json=mock_asset
    )
    asset_page = client.get(
        url_for("AssetCrudUI:get", id=asset.id), follow_redirects=True
    )
    assert ("Edit asset %s" % mock_asset["display_name"]).encode() in asset_page.data
    assert str(mock_asset["capacity_in_mw"]).encode() in asset_page.data
    assert str(mock_asset["latitude"]).encode() in asset_page.data
    assert str(mock_asset["longitude"]).encode() in asset_page.data


def test_edit_asset(db, client, setup_assets, requests_mock, as_admin):
    mock_asset = mock_asset_response(as_list=False)
    requests_mock.patch(
        "http://localhost//api/v2_0/asset/1", status_code=200, json=mock_asset
    )
    response = client.post(
        url_for("AssetCrudUI:post", id=1),
        follow_redirects=True,
        data=mock_api_data_as_form_input(mock_asset),
    )
    assert response.status_code == 200
    assert b"Editing was successful" in response.data
    assert mock_asset["display_name"] in str(response.data)
    assert str(mock_asset["latitude"]) in str(response.data)
    assert str(mock_asset["longitude"]) in str(response.data)


def test_add_asset(db, client, setup_assets, requests_mock, as_admin):
    """Add a new asset"""
    prosumer = find_user_by_email("test_prosumer@seita.nl")
    mock_asset = mock_asset_response(owner_id=prosumer.id, as_list=False)
    requests_mock.post(
        "http://localhost//api/v2_0/assets", status_code=201, json=mock_asset
    )
    response = client.post(
        url_for("AssetCrudUI:post", id="create"),
        follow_redirects=True,
        data=mock_api_data_as_form_input(mock_asset),
    )
    assert response.status_code == 200  # response is HTML form
    assert "html" in response.content_type
    assert b"Creation was successful" in response.data
    assert mock_asset["display_name"] in str(response.data)
    assert str(mock_asset["latitude"]) in str(response.data)
    assert str(mock_asset["longitude"]) in str(response.data)


def test_delete_asset(client, db, requests_mock, as_admin):
    """Delete an asset"""
    requests_mock.delete("http://localhost//api/v2_0/asset/1", status_code=204, json={})
    requests_mock.get("http://localhost//api/v2_0/assets", status_code=200, json={})
    response = client.get(
        url_for("AssetCrudUI:delete_with_data", id=1),
        follow_redirects=True,
    )
    assert response.status_code == 200
    assert b"have been deleted" in response.data



from datetime import datetime, timedelta
import pytest

import numpy as np
import pandas as pd

from flexmeasures.data.models.assets import Asset
from flexmeasures.data.models.markets import Market
from flexmeasures.data.models.planning.battery import schedule_battery
from flexmeasures.data.models.planning.charging_station import schedule_charging_station
from flexmeasures.utils.calculations import integrate_time_series
from flexmeasures.utils.time_utils import as_server_time


def test_battery_solver_day_1(add_battery_assets):
    epex_da = Market.query.filter(Market.name == "epex_da").one_or_none()
    battery = Asset.query.filter(Asset.name == "Test battery").one_or_none()
    start = as_server_time(datetime(2015, 1, 1))
    end = as_server_time(datetime(2015, 1, 2))
    resolution = timedelta(minutes=15)
    soc_at_start = battery.soc_in_mwh
    schedule = schedule_battery(battery, epex_da, start, end, resolution, soc_at_start)
    soc_schedule = integrate_time_series(schedule, soc_at_start, decimal_precision=6)

    with pd.option_context("display.max_rows", None, "display.max_columns", 3):
        print(soc_schedule)

    # Check if constraints were met
    assert min(schedule.values) >= battery.capacity_in_mw * -1
    assert max(schedule.values) <= battery.capacity_in_mw
    for soc in soc_schedule.values:
        assert soc >= battery.min_soc_in_mwh
        assert soc <= battery.max_soc_in_mwh


def test_battery_solver_day_2(add_battery_assets):
    epex_da = Market.query.filter(Market.name == "epex_da").one_or_none()
    battery = Asset.query.filter(Asset.name == "Test battery").one_or_none()
    start = as_server_time(datetime(2015, 1, 2))
    end = as_server_time(datetime(2015, 1, 3))
    resolution = timedelta(minutes=15)
    soc_at_start = battery.soc_in_mwh
    schedule = schedule_battery(battery, epex_da, start, end, resolution, soc_at_start)
    soc_schedule = integrate_time_series(schedule, soc_at_start, decimal_precision=6)

    with pd.option_context("display.max_rows", None, "display.max_columns", 3):
        print(soc_schedule)

    # Check if constraints were met
    assert min(schedule.values) >= battery.capacity_in_mw * -1
    assert max(schedule.values) <= battery.capacity_in_mw
    for soc in soc_schedule.values:
        assert soc >= battery.min_soc_in_mwh
        assert soc <= battery.max_soc_in_mwh

    # Check whether the resulting soc schedule follows our expectations for 8 expensive, 8 cheap and 8 expensive hours
    assert (
        soc_schedule.iloc[-1] == battery.min_soc_in_mwh
    )  # Battery sold out at the end of its planning horizon
    assert (
        soc_schedule.loc[start + timedelta(hours=8)] == battery.min_soc_in_mwh
    )  # Sell what you begin with
    assert (
        soc_schedule.loc[start + timedelta(hours=16)] == battery.max_soc_in_mwh
    )  # Buy what you can to sell later


@pytest.mark.parametrize(
    "target_soc, charging_station_name",
    [
        (1, "Test charging station"),
        (5, "Test charging station"),
        (0, "Test charging station (bidirectional)"),
        (5, "Test charging station (bidirectional)"),
    ],
)
def test_charging_station_solver_day_2(target_soc, charging_station_name):
    """Starting with a state of charge 1 kWh, within 2 hours we should be able to reach
    any state of charge in the range [1, 5] kWh for a unidirectional station,
    or [0, 5] for a bidirectional station."""
    soc_at_start = 1
    duration_until_target = timedelta(hours=2)

    epex_da = Market.query.filter(Market.name == "epex_da").one_or_none()
    charging_station = Asset.query.filter(
        Asset.name == charging_station_name
    ).one_or_none()
    start = as_server_time(datetime(2015, 1, 2))
    end = as_server_time(datetime(2015, 1, 3))
    resolution = timedelta(minutes=15)
    target_soc_datetime = start + duration_until_target
    soc_targets = pd.Series(
        np.nan, index=pd.date_range(start, end, freq=resolution, closed="right")
    )
    soc_targets.loc[target_soc_datetime] = target_soc
    consumption_schedule = schedule_charging_station(
        charging_station, epex_da, start, end, resolution, soc_at_start, soc_targets
    )
    soc_schedule = integrate_time_series(
        consumption_schedule, soc_at_start, decimal_precision=6
    )

    # Check if constraints were met
    assert min(consumption_schedule.values) >= charging_station.capacity_in_mw * -1
    assert max(consumption_schedule.values) <= charging_station.capacity_in_mw
    print(consumption_schedule.head(12))
    print(soc_schedule.head(12))
    assert abs(soc_schedule.loc[target_soc_datetime] - target_soc) < 0.00001



"""merge

Revision ID: 31f251554682
Revises: b2b43f0eec40, 61bfc6e45c4d
Create Date: 2018-09-20 12:05:36.768426

"""


# revision identifiers, used by Alembic.
revision = "31f251554682"
down_revision = ("b2b43f0eec40", "61bfc6e45c4d")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import os

import click

if os.name == "nt":
    from rq_win import WindowsWorker as SimpleWorker
else:
    from rq import SimpleWorker


def work_on_rq(redis_queue, exc_handler=None):
    exc_handlers = []
    if exc_handler is not None:
        exc_handlers.append(exc_handler)
    print("STARTING SIMPLE RQ WORKER, seeing %d job(s)" % redis_queue.count)
    worker = SimpleWorker(
        [redis_queue],
        connection=redis_queue.connection,
        exception_handlers=exc_handlers,
    )
    worker.work(burst=True)


def exception_reporter(job, exc_type, exc_value, traceback):
    click.echo("HANDLING RQ WORKER EXCEPTION: %s:%s\n" % (exc_type, exc_value))



"""create_measurement_table

Revision ID: 45d937300b0f
Revises: a328412b4623
Create Date: 2018-05-07 18:18:59.555454

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "45d937300b0f"
down_revision = "a328412b4623"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "measurement",
        sa.Column("datetime", sa.DateTime(timezone=True), nullable=False),
        sa.Column("asset_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Float(), nullable=False),
        sa.ForeignKeyConstraint(["asset_id"], ["asset.id"]),
        sa.PrimaryKeyConstraint("datetime", "asset_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("measurement")
    # ### end Alembic commands ###



from typing import List
from datetime import timedelta

import isodate
import pytest

from flask_security.utils import hash_password

from flexmeasures.data.services.users import create_user


@pytest.fixture(scope="module", autouse=True)
def setup_api_test_data(db, setup_account, setup_roles_users, add_market_prices):
    """
    Set up data for API v1 tests.
    """
    print("Setting up data for API v1 tests on %s" % db.engine)

    from flexmeasures.data.models.assets import Asset, AssetType, Power
    from flexmeasures.data.models.data_sources import DataSource

    # Create an anonymous user
    test_anonymous_prosumer = create_user(
        username="anonymous user with Prosumer role",
        email="demo@seita.nl",
        password=hash_password("testtest"),
        account_name=setup_account.name,
        user_roles=[
            "Prosumer",
            dict(name="anonymous", description="Anonymous test user"),
        ],
    )

    # Create 1 test asset for the anonymous user
    test_asset_type = AssetType(name="test-type")
    db.session.add(test_asset_type)
    asset_names = ["CS 0"]
    assets: List[Asset] = []
    for asset_name in asset_names:
        asset = Asset(
            name=asset_name,
            asset_type_name="test-type",
            event_resolution=timedelta(minutes=15),
            capacity_in_mw=1,
            latitude=100,
            longitude=100,
            unit="MW",
        )
        asset.owner = test_anonymous_prosumer
        assets.append(asset)
        db.session.add(asset)

    # Create a test user without a USEF role
    create_user(
        username="test user without roles",
        email="test_user@seita.nl",
        password=hash_password("testtest"),
        account_name=setup_account.name,
    )

    # Create 5 test assets for the test_prosumer user
    test_prosumer = setup_roles_users["Test Prosumer"]
    asset_names = ["CS 1", "CS 2", "CS 3", "CS 4", "CS 5"]
    assets: List[Asset] = []
    for asset_name in asset_names:
        asset = Asset(
            name=asset_name,
            asset_type_name="test-type",
            event_resolution=timedelta(minutes=15),
            capacity_in_mw=1,
            latitude=100,
            longitude=100,
            unit="MW",
        )
        asset.owner = test_prosumer
        if asset_name == "CS 4":
            asset.event_resolution = timedelta(hours=1)
        assets.append(asset)
        db.session.add(asset)

    # Add power forecasts to one of the assets, for two sources
    cs_5 = Asset.query.filter(Asset.name == "CS 5").one_or_none()
    test_supplier = setup_roles_users["Test Supplier"]
    prosumer_data_source = DataSource.query.filter(
        DataSource.user == test_prosumer
    ).one_or_none()
    supplier_data_source = DataSource.query.filter(
        DataSource.user == test_supplier
    ).one_or_none()
    meter_data = []
    for i in range(6):
        p_1 = Power(
            datetime=isodate.parse_datetime("2015-01-01T00:00:00Z")
            + timedelta(minutes=15 * i),
            horizon=timedelta(0),
            value=(100.0 + i) * -1,
            asset_id=cs_5.id,
            data_source_id=prosumer_data_source.id,
        )
        p_2 = Power(
            datetime=isodate.parse_datetime("2015-01-01T00:00:00Z")
            + timedelta(minutes=15 * i),
            horizon=timedelta(hours=0),
            value=(1000.0 - 10 * i) * -1,
            asset_id=cs_5.id,
            data_source_id=supplier_data_source.id,
        )
        meter_data.append(p_1)
        meter_data.append(p_2)
    db.session.bulk_save_objects(meter_data)

    print("Done setting up data for API v1 tests")


@pytest.fixture(scope="function")
def setup_fresh_api_test_data(fresh_db, setup_roles_users_fresh_db):
    db = fresh_db
    setup_roles_users = setup_roles_users_fresh_db
    from flexmeasures.data.models.assets import Asset, AssetType

    # Create 5 test assets for the test_prosumer user
    test_prosumer = setup_roles_users["Test Prosumer"]
    test_asset_type = AssetType(name="test-type")
    db.session.add(test_asset_type)
    asset_names = ["CS 1", "CS 2", "CS 3", "CS 4", "CS 5"]
    assets: List[Asset] = []
    for asset_name in asset_names:
        asset = Asset(
            name=asset_name,
            asset_type_name="test-type",
            event_resolution=timedelta(minutes=15),
            capacity_in_mw=1,
            latitude=100,
            longitude=100,
            unit="MW",
        )
        asset.owner = test_prosumer
        if asset_name == "CS 4":
            asset.event_resolution = timedelta(hours=1)
        assets.append(asset)
        db.session.add(asset)



"""create table for timed beliefs

Revision ID: e62ac5f519d7
Revises: a528c3c81506
Create Date: 2021-03-28 16:26:45.025994

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "e62ac5f519d7"
down_revision = "a528c3c81506"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "timed_belief",
        sa.Column(
            "event_start", sa.DateTime(timezone=True), nullable=False, index=True
        ),
        sa.Column("belief_horizon", sa.Interval(), nullable=False),
        sa.Column("cumulative_probability", sa.Float(), nullable=False, default=0.5),
        sa.Column("event_value", sa.Float(), nullable=False),
        sa.Column("sensor_id", sa.Integer(), nullable=False, index=True),
        sa.Column("source_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["sensor_id"],
            ["sensor.id"],
            name=op.f("timed_belief_sensor_id_sensor_fkey"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["source_id"],
            ["data_source.id"],
            name=op.f("timed_belief_source_id_source_fkey"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "event_start",
            "belief_horizon",
            "cumulative_probability",
            "sensor_id",
            name=op.f("timed_belief_pkey"),
        ),
    )


def downgrade():
    op.drop_table("timed_belief")



from setuptools import setup, find_packages


def load_requirements(use_case):
    reqs = []
    with open("requirements/%s.txt" % use_case, "r") as f:
        reqs = [
            req
            for req in f.read().splitlines()
            if not req.strip() == ""
            and not req.strip().startswith("#")
            and not req.strip().startswith("-c")
        ]
    return reqs


setup(
    name="flexmeasures",
    description="FlexMeasures - A free platform for real-time optimization of flexible energy.",
    author="Seita BV",
    author_email="nicolas@seita.nl",
    url="https://github.com/seitabv/flexmeasures",
    keywords=["smart grid", "renewables", "balancing", "forecasting", "scheduling"],
    python_requires=">=3.7.1",  # not enforced, just info
    install_requires=load_requirements("app"),
    tests_require=load_requirements("test"),
    setup_requires=["pytest-runner", "setuptools_scm"],
    use_scm_version={"local_scheme": "no-local-version"},  # handled by setuptools_scm
    packages=find_packages(),
    include_package_data=True,  # setuptools_scm takes care of adding the files in SCM
    entry_points={
        "console_scripts": [
            "flexmeasures=flexmeasures.utils.app_utils:flexmeasures_cli"
        ],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    long_description="""\
The *FlexMeasures Platform* is a tool for scheduling flexible actions for energy assets.
For this purpose, it performs monitoring, forecasting and scheduling services.

FlexMeasures is fully usable via APIs, which are inspired by the Universal Smart Energy Framework (USEF).
Some algorithms are included, but projects will usually write their own (WIP).

Energy Flexibility is one of the key ingredients to reducing CO2. FlexMeasures is meant
to facilitate the transition to a carbon-free energy system. By open-sourcing FlexMeasures,
we hope to speed up this transition world-wide.

Please visit https://flexmeasures.io to learn more.
""",
)



from datetime import timedelta
from typing import Optional

import click
from flask import current_app as app
from flask.cli import with_appcontext
from flask_mail import Message
from sentry_sdk import (
    capture_message as capture_message_for_sentry,
    set_context as set_sentry_context,
)

from flexmeasures.data.models.task_runs import LatestTaskRun
from flexmeasures.utils.time_utils import server_now


@click.group("monitor")
def fm_monitor():
    """FlexMeasures: Monitor tasks."""


def send_monitoring_alert(
    task_name: str, msg: str, latest_run: Optional[LatestTaskRun] = None
):
    """
    Send any monitoring message per Sentry and per email. Also log an error.
    """
    latest_run_txt = ""
    if latest_run:
        set_sentry_context(
            "latest_run", {"time": latest_run.datetime, "status": latest_run.status}
        )
        latest_run_txt = (
            f"Last run was at {latest_run.datetime}, status was: {latest_run.status}"
        )

    capture_message_for_sentry(msg)

    email_recipients = app.config.get("FLEXMEASURES_MONITORING_MAIL_RECIPIENTS", [])
    if len(email_recipients) > 0:
        email = Message(subject=f"Problem with task {task_name}", bcc=email_recipients)
        email.body = f"{msg}\n\n{latest_run_txt}\nWe suggest to check the logs."
        app.mail.send(email)

    app.logger.error(f"{msg} {latest_run_txt}")


@fm_monitor.command("tasks")
@with_appcontext
@click.option(
    "--task",
    type=(str, int),
    multiple=True,
    required=True,
    help="The name of the task and the maximal allowed minutes between successful runs. Use multiple times if needed.",
)
def monitor_tasks(task):
    """
    Check if the given task's last successful execution happened less than the allowed time ago.
    If not, alert someone, via email or sentry.
    """
    for t in task:
        task_name = t[0]
        app.logger.info(f"Checking latest run of task {task_name} ...")
        latest_run: LatestTaskRun = LatestTaskRun.query.get(task_name)
        if latest_run is None:
            msg = f"Task {task_name} has no last run and thus cannot be monitored. Is it configured properly?"
            send_monitoring_alert(task_name, msg)
            return
        now = server_now()
        acceptable_interval = timedelta(minutes=t[1])
        # check if latest run was recently enough
        if latest_run.datetime >= now - acceptable_interval:
            # latest run time is okay, let's check the status
            if latest_run.status is False:
                msg = f"A failure has been reported on task {task_name}."
                send_monitoring_alert(task_name, msg, latest_run)
        else:
            msg = (
                f"Task {task_name}'s latest run time is outside of the acceptable range "
                f"({acceptable_interval})."
            )
            app.logger.error(msg)
            send_monitoring_alert(task_name, msg, latest_run)
    app.logger.info("Done checking task runs ...")


app.cli.add_command(fm_monitor)



from typing import Tuple, List, Union
from datetime import datetime, timedelta

from flexmeasures.data.config import db
from flexmeasures.data.models.forecasting.exceptions import NotEnoughDataException
from flexmeasures.utils.time_utils import as_server_time


def check_data_availability(
    generic_asset,
    generic_asset_value_class,
    forecast_start: datetime,
    forecast_end: datetime,
    query_window: Tuple[datetime, datetime],
    horizon: timedelta,
):
    """Check if enough data is available in the database in the first place,
    for training window and lagged variables. Otherwise, suggest new forecast period.
    TODO: we could also check regressor data, if we get regressor specs passed in here.
    """
    q = generic_asset_value_class.query.join(generic_asset.__class__).filter(
        generic_asset.__class__.name == generic_asset.name
    )
    first_value = q.order_by(generic_asset_value_class.datetime.asc()).first()
    last_value = q.order_by(generic_asset_value_class.datetime.desc()).first()
    if first_value is None:
        raise NotEnoughDataException(
            "No data available at all. Forecasting impossible."
        )
    first = as_server_time(first_value.datetime)
    last = as_server_time(last_value.datetime)
    if query_window[0] < first:
        suggested_start = forecast_start + (first - query_window[0])
        raise NotEnoughDataException(
            f"Not enough data to forecast {generic_asset.name} "
            f"for the forecast window {as_server_time(forecast_start)} to {as_server_time(forecast_end)}. "
            f"I needed to query from {as_server_time(query_window[0])}, "
            f"but the first value available is from {first} to {first + generic_asset.event_resolution}. "
            f"Consider setting the start date to {as_server_time(suggested_start)}."
        )
    if query_window[1] - horizon > last + generic_asset.event_resolution:
        suggested_end = forecast_end + (last - (query_window[1] - horizon))
        raise NotEnoughDataException(
            f"Not enough data to forecast {generic_asset.name} "
            f"for the forecast window {as_server_time(forecast_start)} to {as_server_time(forecast_end)}. "
            f"I needed to query until {as_server_time(query_window[1] - horizon)}, "
            f"but the last value available is from {last} to {last + generic_asset.event_resolution}. "
            f"Consider setting the end date to {as_server_time(suggested_end)}."
        )


def create_lags(
    n_lags: int,
    generic_asset_type: db.Model,
    horizon: timedelta,
    resolution: timedelta,
    use_periodicity: bool,
) -> List[timedelta]:
    """ List the lags for this asset type, using horizon and resolution information."""
    lags = []

    # Include a zero lag in case of backwards forecasting
    # Todo: we should always take into account the latest forecast, so always append the zero lag if that belief exists
    if horizon < timedelta(hours=0):
        lags.append(timedelta(hours=0))

    # Include latest measurements
    lag_period = resolution
    number_of_nan_lags = 1 + (horizon - resolution) // lag_period
    for L in range(n_lags):
        lags.append((L + number_of_nan_lags) * lag_period)

    # Include relevant measurements given the asset's periodicity
    if use_periodicity and hasattr(generic_asset_type, "daily_seasonality"):
        if generic_asset_type.daily_seasonality:
            lag_period = timedelta(days=1)
            number_of_nan_lags = 1 + (horizon - resolution) // lag_period
            for L in range(n_lags):
                lags.append((L + number_of_nan_lags) * lag_period)

    # Remove possible double entries
    return list(set(lags))


def get_query_window(
    training_start: datetime, forecast_end: datetime, lags: List[timedelta]
) -> Tuple[datetime, datetime]:
    """Derive query window from start and end date, as well as lags (if any).
    This makes sure we have enough data for lagging and forecasting."""
    if not lags:
        query_start = training_start
    else:
        query_start = training_start - max(lags)
    query_end = forecast_end
    return query_start, query_end


def set_training_and_testing_dates(
    forecast_start: datetime,
    training_and_testing_period: Union[timedelta, Tuple[datetime, datetime]],
) -> Tuple[datetime, datetime]:
    """If needed (if training_and_testing_period is a timedelta),
    derive training_start and testing_end from forecasting_start,
    otherwise simply return training_and_testing_period.


        |------forecast_horizon/belief_horizon------|
        |                  |-------resolution-------|
        belief_time        event_start      event_end


                           |--resolution--|--resolution--|--resolution--|--resolution--|--resolution--|--resolution--|
        |---------forecast_horizon--------|              |              |              |              |              |
        belief_time        event_start    |              |              |              |              |              |
                       |---------forecast_horizon--------|              |              |              |              |
                       belief_time        event_start    |              |              |              |              |
                           |          |---------forecast_horizon--------|              |              |              |
                           |          belief_time        event_start    |              |              |              |
    |--------max_lag-------|--------training_and_testing_period---------|---------------forecast_period--------------|
    query_start            training_start |              |    testing_end/forecast_start              |   forecast_end
        |------min_lag-----|              |          |---------forecast_horizon--------|              |              |
                           |              |          belief_time        event_start    |              |              |
                           |              |              |          |---------forecast_horizon--------|              |
                           |              |              |          belief_time        event_start    |              |
                           |              |              |              |          |---------forecast_horizon--------|
                           |              |              |              |          belief_time        event_start    |
    |--------------------------------------------------query_window--------------------------------------------------|

    """
    if isinstance(training_and_testing_period, timedelta):
        return forecast_start - training_and_testing_period, forecast_start
    else:
        return training_and_testing_period



from flask_security import roles_accepted
from flask_json import as_json
from bokeh.embed import json_item
from marshmallow import Schema, fields, validate
from webargs.flaskparser import use_args

from flexmeasures.api.v2_0 import flexmeasures_api as flexmeasures_api_v2_0
from flexmeasures.api.v2_0.routes import v2_0_service_listing
from flexmeasures.data.schemas.times import DurationField
from flexmeasures.data.queries.analytics import get_power_data
from flexmeasures.ui.views.analytics import make_power_figure


"""
An endpoint to get a power chart.

This will grow to become code for more charts eventually.
The plan is to separate charts specs from the actual data later,
and to switch to Altair.

For now, we'll keep this endpoint here, with route and implementation in the same file.
When we move forward, we'll review the architecture.
"""


v2_0_service_listing["services"].append(
    {
        "name": "GET /charts/power",
        "access": ["admin", "Prosumer"],
        "description": "Get a Bokeh chart for power data to embed in web pages.",
    },
)


class ChartRequestSchema(Schema):
    """
    This schema describes the request for a chart.
    """

    resource = fields.Str(required=True)
    start_time = fields.DateTime(required=True)
    end_time = fields.DateTime(required=True)
    resolution = DurationField(required=True)
    show_consumption_as_positive = fields.Bool(missing=True)
    show_individual_traces_for = fields.Str(
        missing="none", validate=validate.OneOf(["none", "schedules", "power"])
    )
    forecast_horizon = DurationField(missing="PT6H")


@flexmeasures_api_v2_0.route("/charts/power", methods=["GET"])
@roles_accepted("admin", "Prosumer")
@use_args(ChartRequestSchema(), location="querystring")
@as_json
def get_power_chart(chart_request):
    """API endpoint to get a chart for power data which can be embedded in web pages.

    .. :quickref: Chart; Get a power chart

    This endpoint returns a Bokeh chart with power data which can be embedded in a website.
    It includes forecasts and even schedules, if available.

    **Example request**

    An example of a chart request:

    .. sourcecode:: json

        {
            "resource": ""my-battery,
            "start_time": "2020-02-20:10:00:00UTC",
            "end_time": "2020-02-20:11:00:00UTC",
            "resolution": "PT15M",
            "consumption_as_positive": true
            "resolution": "PT6H",
            "show_individual_traces_for": "none"  // can be power or schedules
        }

    On your webpage, you need to include the Bokeh libraries, e.g.:

        <script src="https://cdn.pydata.org/bokeh/release/bokeh-1.0.4.min.js"></script>

    (The version needs to match the version used by the FlexMeasures server, see requirements/app.txt)

    Then you can call this endpoint and include the result like this:

    .. sourcecode:: javascript

        <script>
            fetch('http://localhost:5000/api/v2_0/charts/power?' + urlData.toString(),
            {
                method: "GET",
                mode: "cors",
                headers:
                    {
                        "Content-Type": "application/json",
                        "Authorization": "<users auth token>"
                    },
            })
            .then(function(response) { return response.json(); })
            .then(function(item) { Bokeh.embed.embed_item(item, "<ID of the div >"); });
        </script>

    where `urlData` is a `URLSearchData` object and contains the chart request parameters (see above).

    :reqheader Authorization: The authentication token
    :reqheader Content-Type: application/json
    :resheader Content-Type: application/json
    :status 200: PROCESSED
    :status 400: INVALID_REQUEST
    :status 401: UNAUTHORIZED
    :status 403: INVALID_SENDER
    :status 422: UNPROCESSABLE_ENTITY
    """
    data = get_power_data(
        resource=chart_request["resource"],
        show_consumption_as_positive=chart_request["show_consumption_as_positive"],
        showing_individual_traces_for=chart_request["show_individual_traces_for"],
        metrics={},  # will be stored here, we don't need them for now
        query_window=(chart_request["start_time"], chart_request["end_time"]),
        resolution=chart_request["resolution"],
        forecast_horizon=chart_request["forecast_horizon"],
    )
    figure = make_power_figure(
        resource_display_name=chart_request["resource"],
        data=data[0],
        forecast_data=data[1],
        schedule_data=data[2],
        show_consumption_as_positive=chart_request["show_consumption_as_positive"],
        shared_x_range=None,
        sizing_mode="scale_both",
    )
    return json_item(figure)



"""create_price_table

Revision ID: 1e8d27922f56
Revises: 9254559dcac2
Create Date: 2018-05-28 13:30:54.227839

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "1e8d27922f56"
down_revision = "9254559dcac2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "price",
        sa.Column("datetime", sa.DateTime(timezone=True), nullable=False),
        sa.Column("market_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Float(), nullable=False),
        sa.ForeignKeyConstraint(["market_id"], ["market.id"]),
        sa.PrimaryKeyConstraint("datetime", "market_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("price")
    # ### end Alembic commands ###



"""drop forecasting jobs table and make display names nullable

Revision ID: 7113b0f00678
Revises: ac2613fffc74
Create Date: 2020-06-04 11:29:46.507095

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "7113b0f00678"
down_revision = "ac2613fffc74"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("forecasting_job")
    op.alter_column(
        "asset_type",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=True,
        existing_server_default=sa.text("''::character varying"),
    )
    op.alter_column(
        "market",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=True,
        existing_server_default=sa.text("''::character varying"),
    )
    op.alter_column(
        "market_type",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=True,
        existing_server_default=sa.text("''::character varying"),
    )
    op.alter_column(
        "weather_sensor_type",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=True,
        existing_server_default=sa.text("''::character varying"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "weather_sensor_type",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=False,
        existing_server_default=sa.text("''::character varying"),
    )
    op.alter_column(
        "market_type",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=False,
        existing_server_default=sa.text("''::character varying"),
    )
    op.alter_column(
        "market",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=False,
        existing_server_default=sa.text("''::character varying"),
    )
    op.alter_column(
        "asset_type",
        "display_name",
        existing_type=sa.VARCHAR(length=80),
        nullable=False,
        existing_server_default=sa.text("''::character varying"),
    )
    op.create_table(
        "forecasting_job",
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column(
            "timed_value_type",
            sa.VARCHAR(length=30),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column("asset_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column(
            "start",
            postgresql.TIMESTAMP(timezone=True),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            "end",
            postgresql.TIMESTAMP(timezone=True),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            "horizon", postgresql.INTERVAL(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "in_progress_since",
            postgresql.TIMESTAMP(timezone=True),
            autoincrement=False,
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id", name="forecasting_job_pkey"),
    )
    # ### end Alembic commands ###



from flask import Flask
from flask_security import auth_token_required, roles_accepted


def register_at(app: Flask):
    """This can be used to register FlaskViews."""

    from flexmeasures.api.dev.sensors import SensorAPI
    from flexmeasures.api.dev.sensor_data import post_data as post_sensor_data_impl

    SensorAPI.register(app, route_prefix="/api/dev")

    @app.route("/sensorData", methods=["POST"])
    @auth_token_required
    @roles_accepted("admin", "MDC", "Prosumer")
    def post_sensor_data():
        """
        Post sensor data to FlexMeasures.

        For example:

        {
            "type": "PostSensorDataRequest",
            "sensor": "ea1.2021-01.io.flexmeasures:fm1.1",
            "values": [-11.28, -11.28, -11.28, -11.28],
            "start": "2021-06-07T00:00:00+02:00",
            "duration": "PT1H",
            "unit": "m³/h",
        }

        The above request posts four values for a duration of one hour, where the first
        event start is at the given start time, and subsequent values start in 15 minute intervals throughout the one hour duration.

        The sensor is the one with ID=1.
        The unit has to match the sensor's required unit.
        The resolution of the data has to match the sensor's required resolution, but
        FlexMeasures will attempt to upsample lower resolutions.
        """
        return post_sensor_data_impl()

    # TODO: add GET /sensorData



from datetime import datetime, timedelta
from typing import List, Union, Tuple, Optional

from flask import current_app
from flask_security.core import current_user
from humanize import naturaldate, naturaltime
import pandas as pd
from pandas.tseries.frequencies import to_offset
import pytz
import tzlocal
from dateutil import tz


def server_now() -> datetime:
    """The current time of the FlexMeasures platform. UTC time, localized to the FlexMeasures timezone."""
    return as_server_time(datetime.utcnow())


def ensure_local_timezone(
    dt: Union[pd.Timestamp, datetime], tz_name: str = "Europe/Amsterdam"
) -> Union[pd.Timestamp, datetime]:
    """If no timezone is given, assume the datetime is in the given timezone and make it explicit.
    Otherwise, if a timezone is given, convert to that timezone."""
    if isinstance(dt, datetime):
        if dt.tzinfo is not None:
            return dt.astimezone(tz.gettz(tz_name))
        else:
            return dt.replace(tzinfo=tz.gettz(tz_name))
    if dt.tzinfo is not None:
        return dt.tz_convert(tz_name)
    else:
        return dt.tz_localize(tz_name)


def as_server_time(dt: datetime) -> datetime:
    """The datetime represented in the timezone of the FlexMeasures platform.
    If dt is naive, we assume it is UTC time.
    """
    return naive_utc_from(dt).replace(tzinfo=pytz.utc).astimezone(get_timezone())


def localized_datetime(dt: datetime) -> datetime:
    """
    Localise a datetime to the timezone of the FlexMeasures platform.
    Note: this will change nothing but the tzinfo field.
    """
    return get_timezone().localize(naive_utc_from(dt))


def naive_utc_from(dt: datetime) -> datetime:
    """
    Return a naive datetime, that is localised to UTC if it has a timezone.
    If dt is naive, we assume it is already in UTC time.
    """
    if not hasattr(dt, "tzinfo") or dt.tzinfo is None:
        # let's hope this is the UTC time you expect
        return dt
    else:
        return dt.astimezone(pytz.utc).replace(tzinfo=None)


def tz_index_naively(
    data: Union[pd.DataFrame, pd.Series, pd.DatetimeIndex]
) -> Union[pd.DataFrame, pd.Series, pd.DatetimeIndex]:
    """Turn any DatetimeIndex into a tz-naive one, then return. Useful for bokeh, for instance."""
    if isinstance(data, pd.DatetimeIndex):
        return data.tz_localize(tz=None)
    if hasattr(data, "index") and isinstance(data.index, pd.DatetimeIndex):
        # TODO: if index is already naive, don't
        data.index = data.index.tz_localize(tz=None)
    return data


def localized_datetime_str(dt: datetime, dt_format: str = "%Y-%m-%d %I:%M %p") -> str:
    """
    Localise a datetime to the timezone of the FlexMeasures platform.
    If no datetime is passed in, use server_now() as basis.

    Hint: This can be set as a jinja filter, so we can display local time in the app, e.g.:
    app.jinja_env.filters['localized_datetime'] = localized_datetime_str
    """
    if dt is None:
        dt = server_now()
    local_tz = get_timezone()
    local_dt = naive_utc_from(dt).astimezone(local_tz)
    return local_dt.strftime(dt_format)


def naturalized_datetime_str(
    dt: Optional[datetime], now: Optional[datetime] = None
) -> str:
    """
    Naturalise a datetime object (into a human-friendly string).
    The dt parameter (as well as the now parameter if you use it)
    can be either naive or tz-aware. We assume UTC in the naive case.

    We use the the humanize library to generate a human-friendly string.
    If dt is not longer ago than 24 hours, we use humanize.naturaltime (e.g. "3 hours ago"),
    otherwise humanize.naturaldate (e.g. "one week ago")

    Hint: This can be set as a jinja filter, so we can display local time in the app, e.g.:
    app.jinja_env.filters['naturalized_datetime'] = naturalized_datetime_str
    """
    if dt is None:
        return "never"
    if now is None:
        now = datetime.utcnow()
    # humanize uses the local now internally, so let's make dt local
    if dt.tzinfo is None:
        local_dt = (
            dt.replace(tzinfo=pytz.utc)
            .astimezone(tzlocal.get_localzone())
            .replace(tzinfo=None)
        )
    else:
        local_dt = dt.astimezone(tzlocal.get_localzone()).replace(tzinfo=None)
    # decide which humanize call to use for naturalization
    if naive_utc_from(dt) >= naive_utc_from(now) - timedelta(hours=24):
        return naturaltime(local_dt)
    else:
        return naturaldate(local_dt)


def resolution_to_hour_factor(resolution: Union[str, timedelta]) -> float:
    """Return the factor with which a value needs to be multiplied in order to get the value per hour,
    e.g. 10 MW at a resolution of 15min are 2.5 MWh per time step.

    :param resolution: timedelta or pandas offset such as "15T" or "1H"
    """
    if isinstance(resolution, timedelta):
        return resolution / timedelta(hours=1)
    return pd.Timedelta(resolution).to_pytimedelta() / timedelta(hours=1)


def decide_resolution(start: Optional[datetime], end: Optional[datetime]) -> str:
    """
    Decide on a practical resolution given the length of the selected time period.
    Useful for querying or plotting.
    """
    if start is None or end is None:
        return "15T"  # default if we cannot tell period
    period_length = end - start
    if period_length > timedelta(weeks=16):
        resolution = "168h"  # So upon switching from days to weeks, you get at least 16 data points
    elif period_length > timedelta(days=14):
        resolution = "24h"  # So upon switching from hours to days, you get at least 14 data points
    elif period_length > timedelta(hours=48):
        resolution = "1h"  # So upon switching from 15min to hours, you get at least 48 data points
    elif period_length > timedelta(hours=8):
        resolution = "15T"
    else:
        resolution = "5T"  # we are (currently) not going lower than 5 minutes
    return resolution


def get_timezone(of_user=False) -> pytz.BaseTzInfo:
    """Return the FlexMeasures timezone, or if desired try to return the timezone of the current user."""
    default_timezone = pytz.timezone(
        current_app.config.get("FLEXMEASURES_TIMEZONE", "")
    )
    if not of_user:
        return default_timezone
    if current_user.is_anonymous:
        return default_timezone
    if current_user.timezone not in pytz.common_timezones:
        return default_timezone
    return pytz.timezone(current_user.timezone)


def round_to_closest_quarter(dt: datetime) -> datetime:
    new_hour = dt.hour
    new_minute = 15 * round((float(dt.minute) + float(dt.second) / 60) / 15)
    if new_minute == 60:
        new_hour += 1
        new_minute = 0
    return datetime(dt.year, dt.month, dt.day, new_hour, new_minute, tzinfo=dt.tzinfo)


def round_to_closest_hour(dt: datetime) -> datetime:
    if dt.minute >= 30:
        return datetime(dt.year, dt.month, dt.day, dt.hour + 1, tzinfo=dt.tzinfo)
    else:
        return datetime(dt.year, dt.month, dt.day, dt.hour, tzinfo=dt.tzinfo)


def get_most_recent_quarter() -> datetime:
    now = server_now()
    return now.replace(minute=now.minute - (now.minute % 15), second=0, microsecond=0)


def get_most_recent_hour() -> datetime:
    now = server_now()
    return now.replace(minute=now.minute - (now.minute % 60), second=0, microsecond=0)


def get_most_recent_clocktime_window(
    window_size_in_minutes: int, now: Optional[datetime] = None
) -> Tuple[datetime, datetime]:
    """
    Calculate a recent time window, returning a start and end minute so that
    a full hour can be filled with such windows, e.g.:

    Calling this function at 15:01:xx with window size 5 -> (14:55:00, 15:00:00)
    Calling this function at 03:36:xx with window size 15 -> (03:15:00, 03:30:00)

    window_size_in_minutes is assumed to > 0 and < = 60, and a divisor of 60 (1, 2, ..., 30, 60).

    If now is not given, the current server time is used.
    if now / the current time lies within a boundary minute (e.g. 15 when window_size_in_minutes=5),
    then the window is not deemed over and the previous one is returned (in this case, [5, 10])

    Returns two datetime objects. They'll be in the timezone (if given) of the now parameter,
    or in the server timezone (see FLEXMEASURES_TIMEZONE setting).
    """
    assert window_size_in_minutes > 0
    assert 60 % window_size_in_minutes == 0
    if now is None:
        now = server_now()
    last_full_minute = now.replace(second=0, microsecond=0) - timedelta(minutes=1)
    last_round_minute = last_full_minute.minute - (
        last_full_minute.minute % window_size_in_minutes
    )
    begin_time = last_full_minute.replace(minute=last_round_minute) - timedelta(
        minutes=window_size_in_minutes
    )
    end_time = begin_time + timedelta(minutes=window_size_in_minutes)
    return begin_time, end_time


def get_default_start_time() -> datetime:
    return get_most_recent_quarter() - timedelta(days=1)


def get_default_end_time() -> datetime:
    return get_most_recent_quarter() + timedelta(days=1)


def get_first_day_of_next_month() -> datetime:
    return (datetime.now().replace(day=1) + timedelta(days=32)).replace(day=1)


def freq_label_to_human_readable_label(freq_label: str) -> str:
    """Translate pandas frequency labels to human-readable labels."""
    f2h_map = {
        "5T": "5 minutes",
        "15T": "15 minutes",
        "1h": "1 hour",
        "24h": "1 day",
        "168h": "1 week",
    }
    return f2h_map.get(freq_label, freq_label)


def forecast_horizons_for(
    resolution: Union[str, timedelta]
) -> Union[List[str], List[timedelta]]:
    """Return a list of horizons that are supported per resolution.
    Return values or of the same type as the input."""
    if isinstance(resolution, timedelta):
        resolution_str = timedelta_to_pandas_freq_str(resolution)
    else:
        resolution_str = resolution
    horizons = []
    if resolution_str in ("5T", "10T"):
        horizons = ["1h", "6h", "24h"]
    elif resolution_str in ("15T", "1h", "H"):
        horizons = ["1h", "6h", "24h", "48h"]
    elif resolution_str in ("24h", "D"):
        horizons = ["24h", "48h"]
    elif resolution_str in ("168h", "7D"):
        horizons = ["168h"]
    if isinstance(resolution, timedelta):
        return [pd.to_timedelta(to_offset(h)) for h in horizons]
    else:
        return horizons


def supported_horizons() -> List[timedelta]:
    return [
        timedelta(hours=1),
        timedelta(hours=6),
        timedelta(hours=24),
        timedelta(hours=48),
    ]


def timedelta_to_pandas_freq_str(resolution: timedelta) -> str:
    return to_offset(resolution).freqstr



from flask import url_for
import pytest

from flexmeasures.data.services.users import find_user_by_email
from flexmeasures.api.tests.utils import get_auth_token, UserContext


@pytest.mark.parametrize("use_auth", [False, True])
def test_get_users_bad_auth(client, use_auth):
    """
    Attempt to get users with insufficient or missing auth.
    """
    # the case without auth: authentication will fail
    headers = {"content-type": "application/json"}
    if use_auth:
        # in this case, we successfully authenticate,
        # but fail authorization (no admin)
        headers["Authorization"] = get_auth_token(
            client, "test_supplier@seita.nl", "testtest"
        )

    get_users_response = client.get(
        url_for("flexmeasures_api_v2_0.get_users"),
        headers=headers,
    )
    print("Server responded with:\n%s" % get_users_response.data)
    if use_auth:
        assert get_users_response.status_code == 403
    else:
        assert get_users_response.status_code == 401


@pytest.mark.parametrize("include_inactive", [False, True])
def test_get_users_inactive(client, setup_inactive_user, include_inactive):
    headers = {
        "content-type": "application/json",
        "Authorization": get_auth_token(client, "test_prosumer@seita.nl", "testtest"),
    }
    query = {}
    if include_inactive in (True, False):
        query["include_inactive"] = include_inactive
    get_users_response = client.get(
        url_for("flexmeasures_api_v2_0.get_users"), query_string=query, headers=headers
    )
    print("Server responded with:\n%s" % get_users_response.json)
    assert get_users_response.status_code == 200
    assert isinstance(get_users_response.json, list)
    if include_inactive is False:
        assert len(get_users_response.json) == 2
    else:
        assert len(get_users_response.json) == 3


def test_get_one_user(client):
    test_supplier_id = find_user_by_email("test_supplier@seita.nl").id
    headers = {
        "content-type": "application/json",
        "Authorization": get_auth_token(client, "test_prosumer@seita.nl", "testtest"),
    }

    get_user_response = client.get(
        url_for("flexmeasures_api_v2_0.get_user", id=test_supplier_id),
        headers=headers,
    )
    print("Server responded with:\n%s" % get_user_response.data)
    assert get_user_response.status_code == 200
    assert get_user_response.json["username"] == "Test Supplier"


def test_edit_user(client):
    with UserContext("test_supplier@seita.nl") as supplier:
        supplier_auth_token = supplier.get_auth_token()  # supplier is no admin
        supplier_id = supplier.id
    with UserContext("test_prosumer@seita.nl") as prosumer:
        prosumer_auth_token = prosumer.get_auth_token()  # prosumer is an admin
        prosumer_id = prosumer.id
    # without being the user themselves or an admin, the user cannot be edited
    user_edit_response = client.patch(
        url_for("flexmeasures_api_v2_0.patch_user", id=prosumer_id),
        headers={
            "content-type": "application/json",
            "Authorization": supplier_auth_token,
        },
        json={},
    )
    assert user_edit_response.status_code == 403
    user_edit_response = client.patch(
        url_for("flexmeasures_api_v2_0.patch_user", id=supplier_id),
        headers={"content-type": "application/json"},
        json={},
    )
    assert user_edit_response.status_code == 401
    # admin can deactivate supplier, other changes will be ignored
    # (id is in the User schema of the API, but we ignore it)
    headers = {"content-type": "application/json", "Authorization": prosumer_auth_token}
    user_edit_response = client.patch(
        url_for("flexmeasures_api_v2_0.patch_user", id=supplier_id),
        headers=headers,
        json={"active": False, "id": 888},
    )
    print("Server responded with:\n%s" % user_edit_response.json)
    assert user_edit_response.status_code == 200
    assert user_edit_response.json["active"] is False
    supplier = find_user_by_email("test_supplier@seita.nl")
    assert supplier.active is False
    assert supplier.id == supplier_id
    # admin can edit themselves but not sensitive fields
    headers = {"content-type": "application/json", "Authorization": prosumer_auth_token}
    user_edit_response = client.patch(
        url_for("flexmeasures_api_v2_0.patch_user", id=prosumer_id),
        headers=headers,
        json={"active": False},
    )
    print("Server responded with:\n%s" % user_edit_response.json)
    assert user_edit_response.status_code == 403


def test_edit_user_with_unexpected_fields(client):
    """Sending unexpected fields (not in Schema) is an Unprocessable Entity error."""
    with UserContext("test_supplier@seita.nl") as supplier:
        supplier_id = supplier.id
    with UserContext("test_prosumer@seita.nl") as prosumer:
        prosumer_auth_token = prosumer.get_auth_token()  # prosumer is an admin
    user_edit_response = client.patch(
        url_for("flexmeasures_api_v2_0.patch_user", id=supplier_id),
        headers={
            "content-type": "application/json",
            "Authorization": prosumer_auth_token,
        },
        json={"active": False, "password": "I-should-not-be-sending-this"},
    )
    print("Server responded with:\n%s" % user_edit_response.json)
    assert user_edit_response.status_code == 422



"""introduce the GenericAssetType table

Revision ID: 565e092a6c5e
Revises: 04f0e2d2924a
Create Date: 2021-07-20 16:16:50.872449

"""
import json

from alembic import context, op
from sqlalchemy import orm
import sqlalchemy as sa

from flexmeasures.data.models.generic_assets import GenericAssetType

# revision identifiers, used by Alembic.
revision = "565e092a6c5e"
down_revision = "04f0e2d2924a"
branch_labels = None
depends_on = None


def upgrade():
    """Add GenericAssetType table

    A GenericAssetType is created for each AssetType, MarketType and WeatherSensorType.
    Optionally, additional GenericAssetTypes can be created using:

    flexmeasures db upgrade +1 -x '{"name": "waste power plant"}' -x '{"name": "EVSE", "description": "Electric Vehicle Supply Equipment"}'

    The +1 makes sure we only upgrade by 1 revision, as these arguments are only meant to be used by this upgrade function.
    """

    upgrade_schema()
    upgrade_data()


def downgrade():
    op.drop_table("generic_asset_type")


def upgrade_data():
    """Data migration adding 1 generic asset type for each user defined generic asset type,
    plus 1 generic asset type for each AssetType, MarketType and WeatherSensorType.
    """

    # Get user defined generic asset types
    generic_asset_types = context.get_x_argument()

    # Declare ORM table views
    t_asset_types = sa.Table(
        "asset_type",
        sa.MetaData(),
        sa.Column("name", sa.String(80)),
        sa.Column("display_name", sa.String(80)),
    )
    t_market_types = sa.Table(
        "market_type",
        sa.MetaData(),
        sa.Column("name", sa.String(80)),
        sa.Column("display_name", sa.String(80)),
    )
    t_weather_sensor_types = sa.Table(
        "weather_sensor_type",
        sa.MetaData(),
        sa.Column("name", sa.String(80)),
        sa.Column("display_name", sa.String(80)),
    )

    # Use SQLAlchemy's connection and transaction to go through the data
    connection = op.get_bind()
    session = orm.Session(bind=connection)

    # Select all existing ids that need migrating, while keeping names intact
    asset_type_results = connection.execute(
        sa.select(
            [
                t_asset_types.c.name,
                t_asset_types.c.display_name,
            ]
        )
    ).fetchall()
    market_type_results = connection.execute(
        sa.select(
            [
                t_market_types.c.name,
                t_market_types.c.display_name,
            ]
        )
    ).fetchall()
    weather_sensor_type_results = connection.execute(
        sa.select(
            [
                t_weather_sensor_types.c.name,
                t_weather_sensor_types.c.display_name,
            ]
        )
    ).fetchall()

    # Prepare to build a list of new generic assets
    new_generic_asset_types = []

    # Construct generic asset type for each user defined generic asset type
    asset_type_results_dict = {k: v for k, v in asset_type_results}
    market_type_results_dict = {k: v for k, v in market_type_results}
    weather_sensor_type_results_dict = {k: v for k, v in weather_sensor_type_results}
    for i, generic_asset_type in enumerate(generic_asset_types):
        generic_asset_type_dict = json.loads(generic_asset_type)
        print(
            f"Constructing one generic asset type according to: {generic_asset_type_dict}"
        )
        if generic_asset_type_dict["name"] in asset_type_results_dict.keys():
            raise ValueError(
                f"User defined generic asset type named '{generic_asset_type_dict['name']}' already exists as asset type."
            )
        if generic_asset_type_dict["name"] in market_type_results_dict.keys():
            raise ValueError(
                f"User defined generic asset type named '{generic_asset_type_dict['name']}' already exists as market type."
            )
        if generic_asset_type_dict["name"] in weather_sensor_type_results_dict.keys():
            raise ValueError(
                f"User defined generic asset type named '{generic_asset_type_dict['name']}' already exists as weather sensor type."
            )
        new_generic_asset_type = GenericAssetType(
            name=generic_asset_type_dict["name"],
            description=generic_asset_type_dict.get("description", None),
        )
        new_generic_asset_types.append(new_generic_asset_type)

    # Construct generic asset types for each AssetType
    if asset_type_results_dict:
        print(
            f"Constructing generic asset types for each of the following asset types: {asset_type_results_dict}"
        )
    for name, display_name in asset_type_results_dict.items():
        # Create new GenericAssets with matching names
        new_generic_asset_type = GenericAssetType(name=name, description=display_name)
        new_generic_asset_types.append(new_generic_asset_type)

    # Construct generic asset types for each MarketType
    if market_type_results_dict:
        print(
            f"Constructing generic asset types for each of the following market types: {market_type_results_dict}"
        )
    for name, display_name in market_type_results_dict.items():
        # Create new GenericAssets with matching names
        new_generic_asset_type = GenericAssetType(name=name, description=display_name)
        new_generic_asset_types.append(new_generic_asset_type)

    # Construct generic asset types for each WeatherSensorType
    if weather_sensor_type_results_dict:
        print(
            f"Constructing generic asset types for each of the following weather sensor types: {weather_sensor_type_results_dict}"
        )
    for name, display_name in weather_sensor_type_results_dict.items():
        # Create new GenericAssets with matching names
        new_generic_asset_type = GenericAssetType(name=name, description=display_name)
        new_generic_asset_types.append(new_generic_asset_type)

    # Add the new generic asset types
    session.add_all(new_generic_asset_types)
    session.commit()


def upgrade_schema():
    op.create_table(
        "generic_asset_type",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=80), nullable=True),
        sa.Column("description", sa.String(length=80), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("generic_asset_type_pkey")),
    )



"""add event_resolution field to Asset, Market & WeatherSennsor

Revision ID: bddc5e9f72a3
Revises: 02ddbbff29a7
Create Date: 2020-10-07 14:12:45.761789

"""
from alembic import op
import sqlalchemy as sa

from datetime import timedelta


# revision identifiers, used by Alembic.
revision = "bddc5e9f72a3"
down_revision = "02ddbbff29a7"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "asset",
        sa.Column(
            "event_resolution",
            sa.Interval(),
            nullable=False,
            server_default=str(timedelta(minutes=0)),
        ),
    )
    op.add_column(
        "market",
        sa.Column(
            "event_resolution",
            sa.Interval(),
            nullable=False,
            server_default=str(timedelta(minutes=0)),
        ),
    )
    op.add_column(
        "weather_sensor",
        sa.Column(
            "event_resolution",
            sa.Interval(),
            nullable=False,
            server_default=str(timedelta(minutes=0)),
        ),
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("weather_sensor", "event_resolution")
    op.drop_column("market", "event_resolution")
    op.drop_column("asset", "event_resolution")
    # ### end Alembic commands ###



"""unique user IDs in data sources

Revision ID: 1a4f0e5c4b86
Revises: 26373d8266db
Create Date: 2018-08-16 12:47:38.492823

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "1a4f0e5c4b86"
down_revision = "26373d8266db"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(None, "data_sources", ["user_id"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "data_sources", type_="unique")
    # ### end Alembic commands ###



from flask import Flask, Blueprint

flexmeasures_api = Blueprint("flexmeasures_api_v2_0", __name__)


def register_at(app: Flask):
    """This can be used to register this blueprint together with other api-related things"""

    import flexmeasures.api.v2_0.routes  # noqa: F401 this is necessary to load the endpoints

    app.register_blueprint(flexmeasures_api, url_prefix="/api/v2_0")



from pathlib import Path

from flask import request, current_app as app
from flask_json import as_json

from flexmeasures.api.common.responses import (
    no_backup,
    request_processed,
    unrecognized_backup,
)
from flexmeasures.data.config import db
from flexmeasures.data.scripts.data_gen import (
    depopulate_measurements,
    depopulate_prognoses,
    depopulate_structure,
    get_affected_classes,
    load_tables,
)


@as_json
def restore_data_response():

    delete_structure = True
    delete_data = True
    restore_structure = True
    restore_data = True

    try:
        backup_name = request.args.get("backup", request.json["backup"])
    except KeyError:
        return no_backup()

    # Make sure backup folder and files exist before restoring
    backup_path = app.config.get("FLEXMEASURES_DB_BACKUP_PATH")
    if (
        not Path("%s/%s" % (backup_path, backup_name)).exists()
        or not Path("%s/%s" % (backup_path, backup_name)).is_dir()
    ):
        return unrecognized_backup()
    affected_classes = get_affected_classes(
        structure=restore_structure, data=restore_data
    )
    for c in affected_classes:
        file_path = "%s/%s/%s.obj" % (backup_path, backup_name, c.__tablename__)
        if not Path(file_path).exists():
            return unrecognized_backup(
                "Can't load table, because filename %s does not exist."
                % c.__tablename__
            )

    # Reset in play mode only (this endpoint should not have been registered otherwise)
    assert app.config.get("FLEXMEASURES_MODE", "") == "play"
    if delete_data:
        depopulate_prognoses(db)
        depopulate_measurements(db)
    if delete_structure:
        depopulate_structure(db)

    # Load backup
    load_tables(
        db,
        backup_name,
        structure=restore_structure,
        data=restore_data,
        backup_path=backup_path,
    )

    return request_processed("Database restored to %s." % backup_name)



import pytest

from marshmallow import ValidationError

from flexmeasures.api.common.schemas.sensor_data import (
    SingleValueField,
    SensorDataSchema,
)


@pytest.mark.parametrize(
    "deserialization_input, exp_deserialization_output",
    [
        (
            1,
            [1],
        ),
        (
            2.7,
            [2.7],
        ),
        (
            [1],
            [1],
        ),
        (
            [2.7],
            [2.7],
        ),
    ],
)
def test_value_field_deserialization(
    deserialization_input,
    exp_deserialization_output,
):
    """Testing straightforward cases"""
    vf = SensorDataSchema.values
    deser = vf.deserialize(deserialization_input)
    assert deser == exp_deserialization_output


@pytest.mark.parametrize(
    "serialization_input, exp_serialization_output",
    [
        (
            1,
            [1],
        ),
        (
            2.7,
            [2.7],
        ),
    ],
)
def test_value_field_serialization(
    serialization_input,
    exp_serialization_output,
):
    """Testing straightforward cases"""
    vf = SingleValueField()
    ser = vf.serialize("values", {"values": serialization_input})
    assert ser == exp_serialization_output


@pytest.mark.parametrize(
    "deserialization_input, error_msg",
    [
        (
            ["three", 4],
            "Not a valid number",
        ),
        (
            "3, 4",
            "Not a valid number",
        ),
    ],
)
def test_value_field_invalid(deserialization_input, error_msg):
    sf = SingleValueField()
    with pytest.raises(ValidationError) as ve:
        sf.deserialize(deserialization_input)
    assert error_msg in str(ve)



import pytest

from flexmeasures.utils.entity_address_utils import (
    build_entity_address,
    parse_entity_address,
    EntityAddressException,
    build_ea_scheme_and_naming_authority,
    reverse_domain_name,
)
from flexmeasures.utils.time_utils import get_first_day_of_next_month


@pytest.mark.parametrize(
    "info, entity_type, host, fm_scheme, exp_result",
    [
        (
            dict(sensor_id=42),
            "sensor",
            "flexmeasures.io",
            "fm1",
            "ea1.2021-01.io.flexmeasures:fm1.42",
        ),
        (
            dict(asset_id=42),
            "sensor",
            "flexmeasures.io",
            "fm1",
            "required field 'sensor_id'",
        ),
        (
            dict(sensor_id=40),
            "connection",
            "flexmeasures.io",
            "fm1",
            "ea1.2021-01.io.flexmeasures:fm1.40",
        ),
        (
            dict(asset_id=40),
            "connection",
            "flexmeasures.io",
            "fm1",
            "required field 'sensor_id'",
        ),
        (
            dict(owner_id=3, asset_id=40),
            "connection",
            "flexmeasures.io",
            "fm0",
            "ea1.2021-01.io.flexmeasures:fm0.3:40",
        ),
        (
            dict(owner_id=3),
            "connection",
            "flexmeasures.io",
            "fm0",
            "required field 'asset_id'",
        ),
        (
            dict(owner_id=40, asset_id=30),
            "connection",
            "localhost:5000",
            "fm0",
            f"ea1.{get_first_day_of_next_month().strftime('%Y-%m')}.localhost:fm0.40:30",
        ),
        (
            dict(
                weather_sensor_type_name="temperature",
                latitude=52,
                longitude=73.0,
            ),
            "weather_sensor",
            "flexmeasures.io",
            "fm0",
            "ea1.2021-01.io.flexmeasures:fm0.temperature:52:73.0",
        ),
        (
            dict(market_name="epex_da"),
            "market",
            "flexmeasures.io",
            "fm0",
            "ea1.2021-01.io.flexmeasures:fm0.epex_da",
        ),
        (
            dict(
                owner_id=40,
                asset_id=30,
                event_type="soc",
                event_id=302,
            ),
            "event",
            "http://staging.flexmeasures.io:4444",
            "fm0",
            "ea1.2022-09.io.flexmeasures.staging:fm0.40:30:302:soc",
        ),
    ],
)
def test_build_entity_address(
    app, info: dict, entity_type: str, host: str, fm_scheme: str, exp_result: str
):
    with app.app_context():
        app.config["FLEXMEASURES_HOSTS_AND_AUTH_START"] = {
            "flexmeasures.io": "2021-01",
            "staging.flexmeasures.io": "2022-09",
        }
        if exp_result.startswith("ea1"):
            assert (
                build_entity_address(info, entity_type, host, fm_scheme) == exp_result
            )
        else:
            with pytest.raises(EntityAddressException, match=exp_result):
                build_entity_address(info, entity_type, host, fm_scheme) == exp_result


@pytest.mark.parametrize(
    "entity_type, entity_address, exp_result",
    [
        (
            "connection",
            "ea2.2018-06.localhost:40:30",
            "starts with 'ea1'",
        ),
        (
            "connection",
            "ea1.2018-RR.localhost:40:30",
            "date specification",
        ),
        (
            "weather_sensor",
            "ea1.2018-04.localhost:5000:40:30",
            "Could not parse",  # no sensor type (which starts with a letter)
        ),
        ("connection", "ea1:40", "a date spec"),  # trying only an asset ID
        (
            "connection",
            "ea1.2018-06.localhost:5000:40:30",
            dict(naming_authority="2018-06.localhost", owner_id=40, asset_id=30),
        ),
        (
            "connection",
            "ea1.2018-06.localhost:40",
            dict(naming_authority="2018-06.localhost", asset_id=40, owner_id=None),
        ),
        (
            "connection",
            "ea1.2018-06.io.flexmeasures:40:30",
            dict(naming_authority="2018-06.io.flexmeasures", owner_id=40, asset_id=30),
        ),
        (
            "weather_sensor",
            "ea1.2018-06.io.flexmeasures:temperature:-52:73.0",
            dict(
                naming_authority="2018-06.io.flexmeasures",
                weather_sensor_type_name="temperature",
                latitude=-52,
                longitude=73.0,
            ),
        ),
        (
            "market",
            "ea1.2018-06.io.flexmeasures:epex_da",
            dict(naming_authority="2018-06.io.flexmeasures", market_name="epex_da"),
        ),
        (
            "event",
            "ea1.2018-06.io.flexmeasures.staging:40:30:302:soc",
            dict(
                naming_authority="2018-06.io.flexmeasures.staging",
                owner_id=40,
                asset_id=30,
                event_type="soc",
                event_id=302,
            ),
        ),
        (
            "event",
            "ea1.2018-06.io.flexmeasures.staging:30:302:soc",
            dict(
                naming_authority="2018-06.io.flexmeasures.staging",
                asset_id=30,
                owner_id=None,
                event_type="soc",
                event_id=302,
            ),
        ),
        (
            "connection",
            "ea1.2018-06.io.flexmeasures.staging:fm1.30",
            dict(
                naming_authority="2018-06.io.flexmeasures.staging",
                asset_id=30,
                owner_id=None,
                event_type="connection",
            ),
        ),
    ],
)
def test_parse_entity_address(entity_type, entity_address, exp_result):
    if isinstance(exp_result, str):  # this means we expect an exception
        with pytest.raises(EntityAddressException, match=exp_result):
            parse_entity_address(
                entity_address, entity_type=entity_type, fm_scheme="fm0"
            )
    else:
        res = parse_entity_address(
            entity_address, entity_type=entity_type, fm_scheme="fm0"
        )
        assert res["scheme"] == "ea1"
        assert res["naming_authority"] == exp_result["naming_authority"]
        if entity_type in ("connection", "event"):
            for field in ("asset_id", "owner_id"):
                assert res[field] == exp_result[field]
        if entity_type == "market":
            assert res["market_name"] == exp_result["market_name"]
        if entity_type == "weather_sensor":
            for field in ("weather_sensor_type_name", "latitude", "longitude"):
                assert res[field] == exp_result[field]
        if entity_type == "event":
            for field in ("event_type", "event_id"):
                assert res[field] == exp_result[field]


def test_reverse_domain_name():
    assert reverse_domain_name("flexmeasures.io") == "io.flexmeasures"
    assert reverse_domain_name("company.flexmeasures.io") == "io.flexmeasures.company"
    assert (
        reverse_domain_name("staging.company.flexmeasures.org")
        == "org.flexmeasures.company.staging"
    )
    assert (
        reverse_domain_name("staging.company.flexmeasures.io:4500")
        == "io.flexmeasures.company.staging"
    )
    assert (
        reverse_domain_name("https://staging.company.flexmeasures.io:4500")
        == "io.flexmeasures.company.staging"
    )
    assert (
        reverse_domain_name("https://user:pass@staging.company.flexmeasures.io:4500")
        == "io.flexmeasures.company.staging"
    )
    assert reverse_domain_name("test.flexmeasures.co.uk") == "uk.co.flexmeasures.test"
    assert (
        reverse_domain_name("test.staging.flexmeasures.co.uk")
        == "uk.co.flexmeasures.staging.test"
    )


def test_build_ea_scheme_and_naming_authority(app):
    assert build_ea_scheme_and_naming_authority(
        "localhost:5000"
    ) == "ea1.%s.localhost" % get_first_day_of_next_month().strftime("%Y-%m")
    assert (
        build_ea_scheme_and_naming_authority("flexmeasures.io")
        == "ea1.2021-01.io.flexmeasures"
    )
    with app.app_context():
        app.config["FLEXMEASURES_HOSTS_AND_AUTH_START"] = {
            "flexmeasures.io": "2021-01",
            "company.flexmeasures.io": "2020-04",
        }
        assert (
            build_ea_scheme_and_naming_authority("company.flexmeasures.io")
            == "ea1.2020-04.io.flexmeasures.company"
        )
    assert (
        build_ea_scheme_and_naming_authority("company.flexmeasures.io", "1999-12")
        == "ea1.1999-12.io.flexmeasures.company"
    )
    with pytest.raises(Exception, match="adhere to the format"):
        build_ea_scheme_and_naming_authority("company.flexmeasures.io", "1999--13")
    with pytest.raises(Exception, match="should be in the range"):
        build_ea_scheme_and_naming_authority("company.flexmeasures.io", "1999-13")
    with pytest.raises(Exception, match="when authority"):
        build_ea_scheme_and_naming_authority("company.notflexmeasures.io")



from flask import Flask, Blueprint

# The api blueprint. It is registered with the Flask app (see app.py)
flexmeasures_api = Blueprint("flexmeasures_api_v1_2", __name__)


def register_at(app: Flask):
    """This can be used to register this blueprint together with other api-related things"""

    import flexmeasures.api.v1_2.routes  # noqa: F401 this is necessary to load the endpoints

    app.register_blueprint(flexmeasures_api, url_prefix="/api/v1_2")



"""create_market_type_and_market_tables

Revision ID: 9254559dcac2
Revises: 45d937300b0f
Create Date: 2018-05-28 13:02:55.210603

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "9254559dcac2"
down_revision = "45d937300b0f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "market_type",
        sa.Column("name", sa.String(length=80), nullable=False),
        sa.Column("daily_seasonality", sa.Boolean(), nullable=False),
        sa.Column("weekly_seasonality", sa.Boolean(), nullable=False),
        sa.Column("yearly_seasonality", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_table(
        "market",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=80), nullable=True),
        sa.Column("market_type_name", sa.String(length=80), nullable=False),
        sa.ForeignKeyConstraint(["market_type_name"], ["market_type.name"]),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("market")
    op.drop_table("market_type")
    # ### end Alembic commands ###



import re
import sys
import traceback
from flask import Flask, jsonify, current_app, request
from werkzeug.exceptions import (
    HTTPException,
    InternalServerError,
    BadRequest,
    NotFound,
)
from sqlalchemy.orm import Query


def log_error(exc: Exception, error_msg: str):
    """Collect meta data about the exception and log it.
    error_msg comes in as an extra attribute because Exception implementations differ here."""
    exc_info = sys.exc_info()
    last_traceback = exc_info[2]

    if hasattr(exc, "__cause__") and exc.__cause__ is not None:
        exc_info = (exc.__cause__.__class__, exc.__cause__, last_traceback)

    extra = dict(url=request.path, **get_err_source_info(last_traceback))

    msg = (
        '{error_name}:"{message}" [occurred at {src_module}({src_func}):{src_linenr},'
        "URL was: {url}]".format(
            error_name=exc.__class__.__name__, message=error_msg, **extra
        )
    )

    current_app.logger.error(msg, exc_info=exc_info)


def get_err_source_info(original_traceback=None) -> dict:
    """Use this when an error is handled to get info on where it occurred."""
    try:  # carefully try to get the actual place where the error happened
        if not original_traceback:
            original_traceback = sys.exc_info()[2]  # class, exc, traceback
        first_call = traceback.extract_tb(original_traceback)[-1]
        return dict(
            src_module=first_call[0],
            src_linenr=first_call[1],
            src_func=first_call[2],
            src_code=first_call[3],
        )
    except Exception as e:
        current_app.warning(
            "I was unable to retrieve error source information: %s." % str(e)
        )
        return dict(module="", linenr=0, method="", src_code="")


def error_handling_router(error: HTTPException):
    """
    Generic handler for errors.
    We respond in json if the request content-type is JSON.
    The ui package can also define how it wants to render HTML errors, by setting a function.
    """
    log_error(error, getattr(error, "description", str(error)))

    http_error_code = 500  # fallback
    if hasattr(error, "code"):
        try:
            http_error_code = int(error.code)
        except ValueError:
            pass

    error_text = getattr(
        error, "description", f"Something went wrong: {error.__class__.__name__}"
    )

    if request.is_json:
        response = jsonify(
            dict(
                message=getattr(error, "description", str(error)),
                status=http_error_code,
            )
        )
        response.status_code = http_error_code
        return response
    # Can UI handle this specific type?
    elif hasattr(current_app, "%s_handler_html" % error.__class__.__name__):
        return getattr(current_app, "%s_handler_html" % error.__class__.__name__)(error)
    # Can UI handle HTTPException? Let's make one from the error.
    elif hasattr(current_app, "HttpException_handler_html"):
        return current_app.HttpException_handler_html(error)
    # This fallback is ugly but better than nothing.
    else:
        return "%s:%s" % (error.__class__.__name__, error_text), http_error_code


def add_basic_error_handlers(app: Flask):
    """Register classes we care about with the generic handler."""
    app.register_error_handler(InternalServerError, error_handling_router)
    app.register_error_handler(BadRequest, error_handling_router)
    app.register_error_handler(HTTPException, error_handling_router)
    app.register_error_handler(NotFound, error_handling_router)
    app.register_error_handler(Exception, error_handling_router)


def print_query(query: Query) -> str:
    """Print full SQLAlchemy query with compiled parameters.

    Recommended use as developer tool only.

    Adapted from https://stackoverflow.com/a/63900851/13775459
    """
    regex = re.compile(r":(?P<name>\w+)")
    params = query.statement.compile().params
    sql = regex.sub(r"'{\g<name>}'", str(query.statement)).format(**params)
    from flexmeasures.data.config import db

    print(f"\nPrinting SQLAlchemy query to database {db.engine.url.database}:\n\n")
    print(sql)
    return sql



from datetime import timedelta
from typing import List, Union

from flask_login import current_user
from marshmallow import fields, post_load, validates_schema, ValidationError
from marshmallow.validate import Equal, OneOf
from marshmallow_polyfield import PolyField
from timely_beliefs import BeliefsDataFrame
import pandas as pd

from flexmeasures.data import ma
from flexmeasures.data.models.data_sources import DataSource
from flexmeasures.api.common.schemas.sensors import SensorField
from flexmeasures.api.common.utils.api_utils import upsample_values
from flexmeasures.data.schemas.times import AwareDateTimeField, DurationField


class SingleValueField(fields.Float):
    """Field that both deserializes and serializes a single value to a list of floats (length 1)."""

    def _deserialize(self, value, attr, obj, **kwargs) -> List[float]:
        return [self._validated(value)]

    def _serialize(self, value, attr, data, **kwargs) -> List[float]:
        return [self._validated(value)]


def select_schema_to_ensure_list_of_floats(
    values: Union[List[float], float], _
) -> Union[fields.List, SingleValueField]:
    """Allows both a single float and a list of floats. Always returns a list of floats.

    Meant to improve user experience by not needing to make a list out of a single item, such that:

        {
            "values": [3.7]
        }

    can be written as:

        {
            "values": 3.7
        }

    Either will be deserialized to [3.7].

    Note that serialization always results in a list of floats.
    This ensures that we are not requiring the same flexibility from users who are retrieving data.
    """
    if isinstance(values, list):
        return fields.List(fields.Float)
    else:
        return SingleValueField()


class SensorDataDescriptionSchema(ma.Schema):
    """
    Describing sensor data (i.e. in a GET request).

    TODO: when we want to support other entity types with this
          schema (assets/weather/markets or actuators), we'll need some re-design.
    """

    type = fields.Str(required=True, validate=Equal("GetSensorDataRequest"))
    sensor = SensorField(required=True, entity_type="sensor", fm_scheme="fm1")
    start = AwareDateTimeField(required=True, format="iso")
    duration = DurationField(required=True)
    horizon = DurationField(
        required=False, missing=timedelta(hours=0), default=timedelta(hours=0)
    )
    unit = fields.Str(required=True)

    @validates_schema
    def check_user_rights_against_sensor(self, data, **kwargs):
        """If the user is a Prosumer and the sensor belongs to an asset
        over which the Prosumer has no ownership, raise a ValidationError.
        """
        # todo: implement check once sensors can belong to an asset
        #       https://github.com/SeitaBV/flexmeasures/issues/155
        pass

    @validates_schema
    def check_schema_unit_against_sensor_unit(self, data, **kwargs):
        # TODO: technically, there are compatible units, like kWh and kW.
        #       They could be allowed here, and the SensorDataSchema could
        #       even convert values to the sensor's unit if possible.
        if data["unit"] != data["sensor"].unit:
            raise ValidationError(
                f"Required unit for this sensor is {data['sensor'].unit}, got: {data['unit']}"
            )


class SensorDataSchema(SensorDataDescriptionSchema):
    """
    This schema includes data, so it can be used for POST requests
    or GET responses.

    TODO: For the GET use case, look at api/common/validators.py::get_data_downsampling_allowed
          (sets a resolution parameter which we can pass to the data collection function).
    """

    type = fields.Str(
        validate=OneOf(["PostSensorDataRequest", "GetSensorDataResponse"])
    )
    values = PolyField(
        deserialization_schema_selector=select_schema_to_ensure_list_of_floats,
        serialization_schema_selector=select_schema_to_ensure_list_of_floats,
        many=False,
    )

    @validates_schema
    def check_resolution_compatibility_of_values(self, data, **kwargs):
        inferred_resolution = data["duration"] / len(data["values"])
        required_resolution = data["sensor"].event_resolution
        # TODO: we don't yet have a good policy w.r.t. zero-resolution (direct measurement)
        if required_resolution == timedelta(hours=0):
            return
        if inferred_resolution % required_resolution != timedelta(hours=0):
            raise ValidationError(
                f"Resolution of {inferred_resolution} is incompatible with the sensor's required resolution of {required_resolution}."
            )

    @post_load()
    def possibly_upsample_values(self, data, **kwargs):
        """
        Upsample the data if needed, to fit to the sensor's resolution.
        Marshmallow runs this after validation.
        """
        inferred_resolution = data["duration"] / len(data["values"])
        required_resolution = data["sensor"].event_resolution

        # TODO: we don't yet have a good policy w.r.t. zero-resolution (direct measurement)
        if required_resolution == timedelta(hours=0):
            return data

        # we already know resolutions are compatible (see validation)
        if inferred_resolution != required_resolution:
            data["values"] = upsample_values(
                data["values"],
                from_resolution=inferred_resolution,
                to_resolution=required_resolution,
            )
        return data

    def load_bdf(sensor_data) -> BeliefsDataFrame:
        """
        Turn the de-serialized and validated data into a BeliefsDataFrame.
        """
        source = DataSource.query.filter(
            DataSource.user_id == current_user.id
        ).one_or_none()
        if not source:
            raise ValidationError(
                f"User {current_user.id} is not an accepted data source."
            )

        num_values = len(sensor_data["values"])
        event_resolution = sensor_data["duration"] / num_values
        dt_index = pd.date_range(
            sensor_data["start"],
            periods=num_values,
            freq=event_resolution,
        )
        s = pd.Series(sensor_data["values"], index=dt_index)
        return BeliefsDataFrame(
            s,
            source=source,
            sensor=sensor_data["sensor"],
            belief_horizon=sensor_data["horizon"],
        )
